import argparse
import logging
import codecs
import logging
from datetime import datetime
import os
import sys
from pathlib import Path

import yaml

from internetnl_domain_analyse import __version__
from internetnl_domain_analyse.domain_analyse_classes import (
    DomainAnalyser,
    DomainPlotter,
    RecordCacheInfo,
)
from internetnl_domain_analyse.domain_plots import (
    make_heatmap,
    make_conditional_score_plot,
    make_conditional_pdf_plot,
    make_verdeling_per_aantal_categorie,
)
from internetnl_domain_analyse.utils import get_windows_or_linux_value

logging.basicConfig(
    format="%(asctime)s %(filename)25s[%(lineno)4s] - %(levelname)-8s : %(message)s",
    level=logging.WARNING,
)
_logger = logging.getLogger()
_log_hc = logging.getLogger("cbsplotlib")
_log_hc.setLevel(_logger.getEffectiveLevel())

MODES = {"statistics", "correlations", "categories", "all"}
IMAGE_TYPES = {"pdf", "png"}


def parse_args():
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(description="Analyse the domains")
    parser.add_argument("settings_filename", help="Settings file")
    parser.add_argument(
        "--version",
        action="version",
        version="{file} version: {ver}".format(
            file=os.path.basename(__file__), ver=__version__
        ),
    )
    parser.add_argument(
        "--quiet",
        dest="loglevel",
        help="set loglevel to WARNING",
        action="store_const",
        const=logging.WARNING,
        default=logging.INFO,
    )
    parser.add_argument(
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
        default=logging.INFO,
    )
    parser.add_argument(
        "--debug",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )
    parser.add_argument("--records_cache_dir", help="Directory of the records cache")
    parser.add_argument("--records_filename", help="Name of the records cache")
    parser.add_argument("--output_filename", help="Name of the output")
    parser.add_argument(
        "--reset", choices={"0", "1"}, default=None, help="Reset the cached data"
    )
    parser.add_argument(
        "--statistics_to_xls",
        help="Write the statistics ot an excel file",
        action="store_true",
    )
    parser.add_argument(
        "--write_dataframe_to_sqlite",
        action="store_true",
        help="Store combined data frame to sqlite and quit",
    )
    parser.add_argument(
        "--show_plots", action="store_true", help="Show each plot before continuing"
    )
    parser.add_argument(
        "--max_plots",
        action="store",
        type=int,
        help="Maximum number of plots. If not given, plot all",
    )
    parser.add_argument("--show_title", action="store_true", help="Show title in plot")
    parser.add_argument("--cumulative", action="store_true", help="Plot pdf cmmulitve")
    parser.add_argument(
        "--not_cumulative",
        action="store_false",
        dest="cumulative",
        help="Do not plot pdf cumulitve",
    )
    parser.add_argument(
        "--plot_all", action="store_true", help="Plot alles", default=False
    )
    parser.add_argument(
        "--cdf_plot", action="store_true", help="Plot de cdf function", default=False
    )
    parser.add_argument(
        "--bar_plot", action="store_true", help="Plot het staafdiagram", default=False
    )
    parser.add_argument(
        "--cate_plot",
        action="store_true",
        help="Plot de heatmap of the categories",
        default=False,
    )
    parser.add_argument(
        "--verdeling_plot",
        action="store_true",
        help="Plot de verdeling per categorie",
        default=False,
    )
    parser.add_argument(
        "--cor_plot", action="store_true", help="Plot de heatmap", default=False
    )
    parser.add_argument(
        "--score_plot",
        action="store_true",
        help="Plot de conditionele score",
        default=False,
    )
    parser.add_argument(
        "--export_highcharts",
        help="Export each image to a highcharts file",
        action="store_true",
    )
    parser.add_argument(
        "--force_plots",
        help="Force making plot, even if it already exists",
        action="store_true",
    )
    parser.add_argument(
        "--latex_files",
        help="Write the latex files containing all de plots",
        action="store_true",
    )
    parser.add_argument(
        "--highcharts_output_directory",
        help="Directory waar alle highcharts naar toe geschreven wordt",
    )
    parser.add_argument(
        "--mode", choices=MODES, default="statistics", help="Type  analyse die we doen"
    )
    parser.add_argument(
        "--bovenschrift",
        action="store_true",
        help="De latex overview file krijgt de captions boven de figuren",
        default=True,
    )
    parser.add_argument(
        "--onderschrift",
        action="store_false",
        dest="bovenschrift",
        help="De latex overview file krijgt de captions boven de figuren",
    )
    parser.add_argument("--scan_type", help="Type van de scan die we willen processen.")
    parser.add_argument(
        "--image_type", choices=IMAGE_TYPES, default="pdf", help="Type van de plaatjes"
    )
    parser.add_argument(
        "--variable_to_plot",
        action="append",
        nargs="*",
        default=None,
        help="Maak alleen het plaatje van deze variabele. "
        "Als niet gegeven dan worden alle variabelen geplot",
    )
    parser.add_argument(
        "--exclude_variable",
        action="append",
        nargs="*",
        default=None,
        help="Sla deze variabele over in het geval alle variabelen geplot worden.",
    )
    parser.add_argument(
        "--statistics",
        action="append",
        nargs="*",
        default=None,
        help="Bereken alleen de breakdowns die gegeven zijn en negeer de "
        "settingsfile flags."
        "Als 'all' in required keys dan doe je alle statistieken",
    )
    parser.add_argument(
        "--plots",
        action="append",
        nargs="*",
        default=None,
        help="Plot alleen de breakdowns die gegeven zijn en negeer de settingsfile "
        "flags."
        "Als 'all' gegeven wordt dan maak je gewoon alle plots",
    )
    parser.add_argument(
        "--tld_extract_cache_directory",
        help="Naam van de directory als je het"
        "script naar cache wilt laten lezen"
        "en schrijven",
    )
    parser.add_argument(
        "--dump_cache_as_sqlite",
        help="Dump de cache files als sqlite zodat je ze in kan zien",
        action="store_true",
    )
    parser.add_argument(
        "--english",
        help="Gebruik Engelse vertaling voor labels van plaatjes",
        action="store_true",
    )
    parsed_arguments = parser.parse_args()

    return parsed_arguments


def set_do_it_vlaggen(required_keys, chapter_info, recursive=False):
    """
    Van een hoofdstukje uit je settings file, druk de do_it vlaggen op

    Args:
        required_keys: list
            List van de items waarvan je de do_it vlag wilt opdrukken
        chapter_info: de dictionary waarvan je de vlaggen zet.
        recursive: bool
            Als dit een recursieve call is, dan willen we de waardes die niet in de lijst zitten
            niet op False zetten

    Returns: dict
        De nieuwe dictionary.
    """
    for key, properties in chapter_info.items():
        if key in required_keys or "all" in required_keys:
            properties["do_it"] = True
            combination: list = properties.get("combination")
            if combination is not None:
                # als de breakdown een combination is dan moet we de onderliggende breakdowns
                # allemaal activeren
                chapter_info = set_do_it_vlaggen(
                    required_keys=combination, chapter_info=chapter_info, recursive=True
                )
        elif not recursive:
            properties["do_it"] = False
    return chapter_info


def main():
    args = parse_args()
    print("-" * 100)
    exe = Path(sys.argv[0]).stem
    now = datetime.now()
    print(
        f"Starting {exe} ({__version__}) at {now} with arguments:\n"
        + " ".join(sys.argv[1:])
    )
    print("-" * 100)

    _logger.setLevel(args.loglevel)

    _logger.info("Reading settings file {}".format(args.settings_filename))
    with codecs.open(args.settings_filename, "r", encoding="UTF-8") as stream:
        settings = yaml.load(stream=stream, Loader=yaml.Loader)

    general_settings = settings["general"]
    cache_directory_base_name = general_settings.get("cache_directory", ".")

    if args.highcharts_output_directory is not None:
        highcharts_directory = args.highcharts_output_directory
    else:
        highcharts_directory = general_settings.get("highcharts_output_directory")

    label_translations = general_settings.get("translations")

    image_directory = Path(general_settings.get("image_directory", "."))
    tex_prepend_path = Path(general_settings.get("tex_prepend_path", "."))
    tex_horizontal_shift = get_windows_or_linux_value(
        general_settings.get("tex_horizontal_shift", "-1.15cm")
    )

    scan_data = general_settings["scan_data"]
    default_scan = general_settings["default_scan"]
    stat_directory = general_settings.get("stat_directory")

    sheet_renames = general_settings["sheet_renames"]
    n_digits = general_settings["n_digits"]
    n_bins = general_settings["n_bins"]
    barh = general_settings.get("barh", False)
    cumulative = general_settings.get("cumulative", False)
    if args.cumulative is not None:
        cumulative = args.cumulative
    show_title = general_settings.get("show_title", False)
    if args.show_title is not None:
        show_title = args.show_title

    bar_plot = args.bar_plot or args.plot_all
    cdf_plot = args.cdf_plot or args.plot_all
    cor_plot = args.cor_plot or args.plot_all
    cate_plot = args.cate_plot or args.plot_all
    verdeling_plot = args.verdeling_plot or args.plot_all
    score_plot = args.score_plot or args.plot_all

    categories = settings.get("categories")
    correlations = settings.get("correlations")
    statistics = settings["statistics"]
    translations = settings["translations"]
    breakdown_labels = settings["breakdown_labels"]
    variables_per_type = settings["variables"]
    module_info = settings["module_info"]
    weights = settings["weight"]
    plot_info = settings["plots"]

    if args.statistics:
        required_keys = [stat[0] for stat in args.statistics]
        statistics = set_do_it_vlaggen(
            required_keys=required_keys, chapter_info=statistics
        )
    if args.plots:
        required_keys = [plot[0] for plot in args.plots]
        plot_info = set_do_it_vlaggen(
            required_keys=required_keys, chapter_info=plot_info
        )

    if args.output_filename is None:
        output_file = general_settings.get("output", "internet_nl_stats")
    else:
        output_file = args.output_filename

    image_directory.mkdir(exist_ok=True)
    var_df = None
    try:
        records_cache_data_per_year = general_settings["records_cache_data"]
    except KeyError as err:
        _logger.warning(err)
        raise KeyError("records cache data needs to be given per year!")

    _logger.info(f"Running domain analyser in {os.getcwd()}")
    for key_scan_type, scan_prop_per_year in scan_data.items():
        if args.scan_type is not None and key_scan_type != args.scan_type:
            _logger.info(f"Scan type {key_scan_type} not requested. Skipping")
            continue

        try:
            variables = variables_per_type[key_scan_type]
        except KeyError as err:
            msg = "Since version in 2022, we need to give the variables per scan type"
            _logger.warning(msg)
            raise KeyError(msg)

        years_to_add_to_plot_legend = list()
        for scan_year, scan_prop in scan_prop_per_year.items():

            if not scan_prop.get("do_it", True):
                continue
            filename = scan_prop["data_file"]
            legend_year_label = scan_prop.get("label", scan_year)
            years_to_add_to_plot_legend.append(legend_year_label)
            if filename is None:
                _logger.info(
                    f"File name for year {scan_year} was None. Skip it but add it to"
                    f"the plot legend"
                )
                continue
            internet_nl_filename = Path(filename)
            records_cache_data = records_cache_data_per_year[scan_year]
            records_cache_info = RecordCacheInfo(
                records_cache_data=records_cache_data,
                year_key=scan_year,
                stat_directory=stat_directory,
            )

            _logger.info(f"Start analyse {scan_year}: {internet_nl_filename}")
            domain_analyses = DomainAnalyser(
                scan_data_key=key_scan_type,
                records_cache_info=records_cache_info,
                internet_nl_filename=internet_nl_filename,
                reset=args.reset,
                output_file=output_file,
                cache_directory_base_name=cache_directory_base_name,
                tld_extract_cache_directory=args.tld_extract_cache_directory,
                statistics=statistics,
                default_scan=default_scan,
                variables=variables,
                module_info=module_info,
                weights=weights,
                translations=translations,
                breakdown_labels=breakdown_labels,
                sheet_renames=sheet_renames,
                n_digits=n_digits,
                write_dataframe_to_sqlite=args.write_dataframe_to_sqlite,
                statistics_to_xls=args.statistics_to_xls,
                n_bins=n_bins,
                mode=args.mode,
                correlations=correlations,
                categories=categories,
                dump_cache_as_sqlite=args.dump_cache_as_sqlite,
            )
            scan_prop["analyses"] = domain_analyses

            if var_df is None:
                var_df = domain_analyses.variables

        if cor_plot and correlations is not None:
            make_heatmap(
                correlations=correlations,
                image_directory=image_directory,
                highcharts_directory=highcharts_directory,
                show_plots=args.show_plots,
                cache_directory=cache_directory_base_name,
            )
        if cate_plot and categories is not None:
            make_conditional_pdf_plot(
                categories=categories,
                image_directory=image_directory,
                highcharts_directory=highcharts_directory,
                show_plots=args.show_plots,
                cache_directory=cache_directory_base_name,
            )
        if verdeling_plot and categories is not None:
            make_verdeling_per_aantal_categorie(
                categories=categories,
                image_directory=image_directory,
                highcharts_directory=highcharts_directory,
                show_plots=args.show_plots,
                cache_directory=cache_directory_base_name,
                export_highcharts=args.export_highcharts,
            )
        if score_plot and correlations is not None:
            make_conditional_score_plot(
                correlations=correlations,
                image_directory=image_directory,
                highcharts_directory=highcharts_directory,
                show_plots=args.show_plots,
                cache_directory=cache_directory_base_name,
            )

        if bar_plot or cdf_plot:
            DomainPlotter(
                scan_data=scan_data,
                scan_data_key=key_scan_type,
                default_scan=default_scan,
                plot_info=plot_info,
                barh=barh,
                show_plots=args.show_plots,
                max_plots=args.max_plots,
                statistics=statistics,
                breakdown_labels=breakdown_labels,
                image_directory=image_directory,
                variables=var_df,
                tex_prepend_path=tex_prepend_path,
                cumulative=cumulative,
                show_title=show_title,
                cdf_plot=cdf_plot,
                bar_plot=bar_plot,
                cache_directory=cache_directory_base_name,
                translations=label_translations,
                export_highcharts=args.export_highcharts,
                highcharts_directory=highcharts_directory,
                tex_horizontal_shift=tex_horizontal_shift,
                bovenschrift=args.bovenschrift,
                image_type=args.image_type,
                variables_to_plot=args.variable_to_plot,
                exclude_variables=args.exclude_variable,
                force_plots=args.force_plots,
                latex_files=args.latex_files,
                years_to_add_to_plot_legend=years_to_add_to_plot_legend,
                module_info=module_info,
                english=args.english,
            )


if __name__ == "__main__":
    main()



import logging
import re
from pathlib import Path

import matplotlib.colors as mpc
import matplotlib.pyplot as plt
import matplotlib.transforms as trn
import numpy as np
import pandas as pd
import seaborn as sns
from cbsplotlib.colors import CBS_COLORS_RBG
from cbsplotlib.highcharts import CBSHighChart
from cbsplotlib.settings import CBSPlotSettings
from cbsplotlib.utils import add_axis_label_background

from internetnl_domain_analyse.utils import get_windows_or_linux_value

_logger = logging.getLogger(__name__)
cbsplotlib_logger = logging.getLogger("cbsplotlib")
cbsplotlib_logger.setLevel(_logger.getEffectiveLevel())
sns.set_style("whitegrid")


class AxisLabel:
    """
    class om de eigenschappen van een as label op te slaan
    """

    def __init__(self, label_properties, text_default=None, positie_default=None):
        self.label_properties = label_properties

        self.text = text_default
        self.positie = positie_default

        self.set_properties()
        if self.text is None:
            raise ValueError(
                "Geen text gezet. Geef label eigenschappen of default waarden mee"
            )
        if self.positie is None:
            raise ValueError(
                "Geen positie gezet. Geef label eigenschappen of default waarden mee"
            )

    def set_properties(self):
        if self.label_properties is not None:
            self.text = self.label_properties["text"]
            if position := self.label_properties.get("positie"):
                self.positie = position


def make_cdf_plot(
    hist,
    grp_key,
    plot_key,
    scan_data_key,
    module_name=None,
    question_name=None,
    image_directory=None,
    show_plots=False,
    figsize=None,
    image_type=None,
    image_file_base=None,
    cummulative=False,
    reference_lines=None,
    xoff=None,
    yoff=None,
    y_max=None,
    y_spacing=None,
    translations=None,
    export_highcharts=None,
    export_svg=False,
    highcharts_info: dict = None,
    title: str = None,
    year: int = None,
    english=False,
):
    figure_properties = CBSPlotSettings()

    if figsize is None:
        figsize = figure_properties.fig_size

    counts = hist[0]
    sum_pdf = counts.sum()
    _logger.info(f"Plot pdf gebaseerd op {sum_pdf} bedrijven (door gewichten)")
    bins = hist[1]
    delta_bin = np.diff(bins)[0]
    pdf = 100 * counts / sum_pdf / delta_bin
    fig, axis = plt.subplots(nrows=1, ncols=1, figsize=figsize)
    fig.subplots_adjust(bottom=0.25, top=0.92, right=0.98)
    axis.tick_params(which="both", bottom=True)

    cdf = pdf.cumsum() * delta_bin

    if cummulative:
        fnc = cdf
        fnc_str = "cdf"
    else:
        fnc = pdf
        fnc_str = "pdf"

    xgrid = bins[:-1] + delta_bin / 2

    axis.set_xlim((0, 100))
    axis.bar(xgrid, fnc, width=delta_bin, edgecolor=None, linewidth=0)

    start, end = axis.get_ylim()
    if y_max is not None:
        end = y_max
    if cummulative:
        axis.yaxis.set_ticks(np.arange(start, end, 25))
    elif y_spacing is not None:
        axis.yaxis.set_ticks(np.arange(start, end + 1, y_spacing))

    if y_max is not None:
        axis.set_ylim((0, y_max))

    stats = dict()
    stats["mean"] = (pdf * bins[:-1]).sum()
    for ii, percentile in enumerate([0, 25, 50, 75, 100]):
        below = cdf < percentile
        if below.all():
            index = cdf.size - 1
        else:
            index = np.argmax(np.diff(cdf < percentile))
        if cummulative:
            pval = fnc[index]
        else:
            if y_max is None:
                pval = end
            else:
                pval = y_max
        value = (index + 1) * delta_bin
        stats[f"p{percentile}"] = value
        _logger.info(f"Adding line {percentile}: {value} {pval}")
        if 0 < percentile < 100:
            axis.vlines(value, 0, pval, color="cbs:appelgroen")
            axis.text(value, 1.02 * pval, f"Q{ii}", color="cbs:appelgroen", ha="center")
    stats_df = pd.DataFrame.from_dict(stats, orient="index", columns=["value"])
    stats_df.index.rename("Stat", inplace=True)

    # this triggers the drawing, otherwise we can not retrieve the xtick labels
    fig.canvas.draw()
    windows_title = f"{grp_key}  {plot_key}"
    try:
        fig.canvas.set_window_title(windows_title)
    except AttributeError as err:
        fig.canvas.manager.set_window_title(windows_title)

    if cummulative:
        if not english:
            y_label = "Cumulatief % van bedrijven met website"
        else:
            y_label = "Cumulative % of companies with met website"

    else:
        if not english:
            y_label = "% van bedrijven met website"
        else:
            y_label = "% of companies with website"

    if translations is not None:
        for key_in, label_out in translations.items():
            if label_out is not None and key_in in y_label:
                _logger.debug(f"Replacing {key_in} -> {label_out}")
                y_label = y_label.replace(key_in, label_out)
            if label_out is not None and key_in in module_name:
                _logger.debug(f"Replacing {key_in} -> {label_out}")
                module_name = module_name.replace(key_in, label_out)

    axis.set_ylabel(y_label, rotation="horizontal", horizontalalignment="left")
    axis.yaxis.set_label_coords(-0.04, 1.05)
    axis.xaxis.grid(False)
    axis.set_xlabel(module_name, horizontalalignment="right")
    axis.xaxis.set_label_coords(0.95, -0.15)
    sns.despine(ax=axis, left=True)

    labels = [_.get_text() for _ in axis.get_xticklabels()]
    axis.xaxis.set_ticks(axis.get_xticks())
    axis.set_xticklabels(labels, ha="center")

    add_axis_label_background(fig=fig, axes=axis, loc="south")

    plot_title = " - ".join([fnc_str, module_name, question_name, plot_key, grp_key])

    image_name_suffix = "_".join([fnc_str, image_file_base, str(year)])
    image_name = "_".join([scan_data_key, plot_key, image_name_suffix])
    image_name_with_ext = ".".join([image_name, image_type])
    image_file = image_directory / Path(image_name_with_ext)

    fig.savefig(image_file)

    stat_file = image_file.with_suffix(".out").as_posix()
    _logger.info(f"Saving stats to {stat_file}")
    stats_df.to_csv(stat_file)

    highcharts_directory = Path(highcharts_info.get("highcharts_directory"))
    highcharts_directory.mkdir(exist_ok=True, parents=True)
    highcharts_label = highcharts_info.get("highcharts_label")

    if export_svg:
        svg_image_file = highcharts_directory / Path(
            "_".join([plot_key, image_name + ".svg"])
        )
        _logger.info(f"Saving plot to {svg_image_file}")
        fig.savefig(svg_image_file)

    if export_highcharts:

        # voor highcharts de titel setten
        if highcharts_label is not None:
            plot_title = highcharts_label
        hc_df = pd.DataFrame(index=bins[:-1], data=fnc, columns=[fnc_str])
        hc_df.index = hc_df.index.rename(module_name)
        CBSHighChart(
            data=hc_df,
            chart_type="column",
            output_directory=highcharts_directory.as_posix(),
            output_file_name=image_file.stem,
            ylabel=y_label,
            title=plot_title,
            enable_legend=False,
        )

    if show_plots:
        plt.show()

    _logger.debug("Done")

    plt.close()

    return image_name_with_ext


def make_bar_plot_horizontal(
    plot_df,
    fig,
    axis,
    margin,
    plot_title,
    show_title,
    translations,
    reference_lines,
    line_iter,
    xoff,
    yoff,
    trans,
    y_spacing_bar_plot,
    y_max_bar_plot,
    legend_position,
    legend_max_columns,
    unit=None,
    english=False,
    bar_width=None,
):
    x_range = None

    if bar_width is not None:
        kwargs = dict(width=bar_width)
    else:
        kwargs = {}

    try:
        plot_df.plot(kind="barh", ax=axis, rot=0, legend=None, **kwargs)
    except IndexError as err:
        _logger.warning(err)
        _logger.warning(f"skip {plot_title}")
        pass
    else:

        # put the high
        axis.invert_yaxis()

        xticks = axis.get_xticks()
        min_x = xticks[0]
        max_x = xticks[-1]
        x_range = max_x - min_x
        if y_max_bar_plot is not None:
            axis.set_xlim((0, y_max_bar_plot))
        else:
            axis.set_xlim((min_x, max_x + 1))
        start, end = axis.get_xlim()
        if y_spacing_bar_plot is not None:
            axis.xaxis.set_ticks(np.arange(start, end + 1, y_spacing_bar_plot))

        if show_title:
            axis.set_title(plot_title)
        axis.set_ylabel("")
        if unit is None:
            if not english:
                x_label = "% van bedrijven met website"
            else:
                x_label = "% of companies with website"
        else:
            x_label = unit

        if translations is not None:
            for key_in, label_out in translations.items():
                if label_out is not None and key_in in x_label:
                    _logger.debug(f"Replacing {key_in} -> {label_out}")
                    x_label = x_label.replace(key_in, label_out)

        axis.set_xlabel(x_label, rotation="horizontal", horizontalalignment="right")
        axis.xaxis.set_label_coords(1.01, -0.12)
        axis.yaxis.grid(False)
        sns.despine(ax=axis, bottom=True)
        axis.tick_params(which="both", left=False)

        add_axis_label_background(
            fig=fig, axes=axis, loc="east", radius_corner_in_mm=1, margin=margin
        )

        number_of_columns = plot_df.columns.values.size
        if legend_max_columns is not None and number_of_columns > legend_max_columns:
            number_of_columns = legend_max_columns
        if legend_position is None:
            legend_bbox_to_anchor = (0.02, 0.00)
        else:
            legend_bbox_to_anchor = legend_position

        legend_bbox_to_anchor = get_windows_or_linux_value(legend_bbox_to_anchor)
        axis.legend(
            loc="lower left",
            frameon=False,
            ncol=number_of_columns,
            bbox_to_anchor=legend_bbox_to_anchor,
            bbox_transform=fig.transFigure,
        )

    if reference_lines is not None:
        color = line_iter.get_next_color()
        for ref_key, ref_line in reference_lines.items():
            ref_label = ref_line["label"]
            ref_plot_df = ref_line["plot_df"]
            value = ref_plot_df.values[0][1]
            color = line_iter.get_next_color()
            axis.axhline(y=value, color=color, linestyle="-.")
            axis.text(
                xoff, value + yoff * x_range, ref_label, color=color, transform=trans
            )


def make_bar_plot_vertical(
    plot_df,
    axis,
    plot_title,
    show_title,
    translations,
    reference_lines,
    line_iter,
    xoff,
    yoff,
    trans,
    unit=None,
    english=False,
):
    y_label = ""
    try:
        plot_df.plot(kind="bar", ax=axis, rot=0, legend=None)
    except IndexError as err:
        _logger.warning(err)
        _logger.warning(f"skip {plot_title}")
        pass
    else:

        yticks = axis.get_yticks()
        min_y = yticks[0]
        max_y = yticks[-1]
        y_range = max_y - min_y
        axis.set_ylim((min_y, max_y))

        if show_title:
            axis.set_title(plot_title)
        axis.set_xlabel("")
        if unit is None:
            if not english:
                x_label = "% van bedrijven met website"
            else:
                x_label = "% of companies with website"
        else:
            x_label = unit

        if translations is not None:
            for key_in, label_out in translations.items():
                if label_out is not None and key_in in y_label:
                    _logger.debug(f"Replacing {key_in} -> {label_out}")
                    y_label = y_label.replace(key_in, label_out)

        axis.set_ylabel(y_label, rotation="horizontal", horizontalalignment="left")
        axis.yaxis.set_label_coords(-0.04, 1.05)
        axis.xaxis.grid(False)
        sns.despine(ax=axis, left=True)
        axis.tick_params(which="both", bottom=False)

        if reference_lines is not None:
            color = line_iter.get_next_color()
            for ref_key, ref_line in reference_lines.items():
                ref_label = ref_line["label"]
                ref_plot_df = ref_line["plot_df"]
                value = ref_plot_df.values[0][1]
                color = line_iter.get_next_color()
                axis.axhline(y=value, color=color, linestyle="-.")
                axis.text(
                    xoff,
                    value + yoff * y_range,
                    ref_label,
                    color=color,
                    transform=trans,
                )


def make_bar_plot(
    plot_df,
    plot_key,
    plot_variable,
    scan_data_key,
    module_name,
    question_name,
    image_directory,
    show_plots=False,
    figsize=None,
    highcharts_height=None,
    image_type="pdf",
    reference_lines=None,
    xoff=0.02,
    yoff=0.02,
    show_title=False,
    barh=False,
    subplot_adjust=None,
    sort_values=False,
    y_max_bar_plot=None,
    y_spacing_bar_plot=None,
    translations=None,
    legend_position=None,
    legend_max_columns=None,
    box_margin=None,
    export_svg=False,
    export_highcharts=False,
    highcharts_directory=None,
    title=None,
    normalize_data=False,
    force_plot=False,
    enable_highcharts_legend=True,
    unit=None,
    english=False,
    bar_width=None,
):
    image_name = re.sub("_\d(\.\d){0,1}$", "", plot_variable)
    image_file = image_directory / Path(
        "_".join([scan_data_key, plot_key, ".".join([image_name, image_type])])
    )
    image_file_name = image_file.as_posix()
    if image_file.exists() and not force_plot:
        _logger.info(f"File {image_file_name} already exists. Skipping plot")
        return image_file_name
    """ create a bar plot from the question 'plot_df'"""
    figure_properties = CBSPlotSettings()

    if figsize is None:
        figsize = figure_properties.fig_size

    _logger.debug(f"Figsize: {figsize}")

    names = plot_df.index.names
    plot_title = " - ".join([module_name, question_name])
    plot_df = plot_df.droplevel(names[:3]).T

    # inverteer de volgorde
    plot_df = plot_df[plot_df.columns[::-1]]
    plot_df = plot_df.reindex(plot_df.index[::-1])

    if normalize_data:
        _logger.info("Normalize data")
        plot_df = 100 * plot_df / plot_df.sum(axis=0)

    fig, axis = plt.subplots(figsize=figsize)
    if subplot_adjust is None:
        s_adjust = dict()
    else:
        s_adjust = subplot_adjust
    bottom = s_adjust.get("bottom", 0.15)
    left = s_adjust.get("left", 0.45)
    top = s_adjust.get("top", 0.95)
    right = s_adjust.get("right", 0.95)
    fig.subplots_adjust(bottom=bottom, left=left, top=top, right=right)

    if box_margin is None:
        margin = 0.1
    else:
        margin = box_margin

    line_iter = axis._get_lines
    trans = trn.blended_transform_factory(axis.transAxes, axis.transData)

    x_label = None
    y_label = None
    y_lim = None

    if not barh:
        make_bar_plot_vertical(
            plot_df=plot_df,
            axis=axis,
            plot_title=plot_title,
            show_title=show_title,
            translations=translations,
            reference_lines=reference_lines,
            line_iter=line_iter,
            xoff=xoff,
            yoff=yoff,
            trans=trans,
            unit=unit,
            english=english,
        )

    else:
        make_bar_plot_horizontal(
            plot_df=plot_df,
            fig=fig,
            axis=axis,
            margin=margin,
            plot_title=plot_title,
            show_title=show_title,
            translations=translations,
            reference_lines=reference_lines,
            line_iter=line_iter,
            xoff=xoff,
            yoff=yoff,
            trans=trans,
            y_spacing_bar_plot=y_spacing_bar_plot,
            y_max_bar_plot=y_max_bar_plot,
            legend_position=legend_position,
            legend_max_columns=legend_max_columns,
            unit=unit,
            english=english,
            bar_width=bar_width,
        )

    _logger.info(f"Saving plot {image_file_name}")
    fig.savefig(image_file)

    if highcharts_directory is not None:
        highcharts_directory.mkdir(exist_ok=True, parents=True)
    if export_svg:
        # met export highcharts gaan we ook een svg exporten
        svg_image_file = highcharts_directory / Path(
            "_".join([plot_key, image_name + ".svg"])
        )
        _logger.info(f"Saving plot {svg_image_file}")
        fig.savefig(svg_image_file)

    if export_highcharts:
        if y_max_bar_plot is not None:
            y_lim = (0, y_max_bar_plot)
        else:
            y_lim = None

        if title is not None:
            plot_title = title
        if barh:
            hc_ylabel = x_label
        else:
            hc_ylabel = y_label
        hc_file = "/".join([highcharts_directory.as_posix(), image_file.stem]) + ".json"
        _logger.info(f"Saving highcharts plot to: {hc_file}")
        CBSHighChart(
            data=plot_df,
            chart_type="bar",
            output_directory=highcharts_directory.as_posix(),
            output_file_name=image_file.stem,
            ylabel=hc_ylabel,
            y_lim=y_lim,
            y_tick_interval=y_spacing_bar_plot,
            title=plot_title,
            enable_legend=enable_highcharts_legend,
            chart_height=highcharts_height,
        )

    if show_plots:
        plt.show()

    plt.close()

    return image_file_name


def make_bar_plot_stacked(
    year,
    plot_df,
    plot_key,
    plot_variable,
    scan_data_key,
    module_name,
    question_name,
    image_directory,
    show_plots=False,
    figsize=None,
    image_type="pdf",
    reference_lines=None,
    xoff=0.02,
    yoff=0.02,
    show_title=False,
    barh=False,
    subplot_adjust=None,
    sort_values=False,
    y_max_bar_plot=None,
    y_spacing_bar_plot=None,
    translations=None,
    legend_position=None,
    box_margin=None,
    export_svg=False,
    export_highcharts=False,
    highcharts_directory=None,
    title=None,
    normalize_data=False,
    force_plot=False,
    enable_highcharts_legend=True,
    unit=None,
    english=False,
):
    image_name = re.sub("_\d(\.\d){0,1}$", "", plot_variable)
    image_name_suffix = "_".join([image_name, str(year)])
    image_name = "_".join([scan_data_key, plot_key, image_name_suffix])
    image_name_with_ext = ".".join([image_name, image_type])
    image_file = image_directory / Path(image_name_with_ext)

    image_file_name = image_file.as_posix()
    if image_file.exists() and not force_plot:
        _logger.info(f"File {image_file_name} already exists. Skipping plot")
        return image_file_name
    """ create a bar plot from the question 'plot_df'"""
    figure_properties = CBSPlotSettings()

    if figsize is None:
        figsize = figure_properties.fig_size

    _logger.debug(f"Figsize: {figsize}")

    names = plot_df.index.names
    plot_title = " - ".join([module_name, question_name])
    plot_df = plot_df.droplevel(names[:3]).T

    # inverteer de volgorde
    # plot_df = plot_df[plot_df.columns[::-1]]
    plot_df = plot_df.reindex(plot_df.index[::-1])

    plot_df.dropna(how="all", axis=0, inplace=True)

    if normalize_data:
        _logger.info("Normalize data")
        plot_df = 100 * plot_df / plot_df.sum(axis=0)

    fig, axis = plt.subplots(figsize=figsize)
    if subplot_adjust is None:
        s_adjust = dict()
    else:
        s_adjust = subplot_adjust
    bottom = s_adjust.get("bottom", 0.15)
    left = s_adjust.get("left", 0.45)
    top = s_adjust.get("top", 0.95)
    right = s_adjust.get("right", 0.95)
    fig.subplots_adjust(bottom=bottom, left=left, top=top, right=right)

    if box_margin is None:
        margin = 0.1
    else:
        margin = box_margin

    line_iter = axis._get_lines
    trans = trn.blended_transform_factory(axis.transAxes, axis.transData)

    x_label = None
    y_label = None
    y_lim = None
    x_range = None

    renames = dict()
    for nr, name in translations.items():
        col = re.sub("_\d(\.\d){0,1}$", f"_{nr}.0", plot_variable)
        renames[col] = name

    _logger.debug(f"Translate with {renames}")

    plot_df.rename(columns=renames, inplace=True)

    try:
        plot_df.plot(kind="barh", ax=axis, rot=0, legend=None, stacked=True)
    except IndexError as err:
        _logger.warning(err)
        _logger.warning(f"skip {plot_title}")
        pass
    else:

        # put the high
        axis.invert_yaxis()

        xticks = axis.get_xticks()
        min_x = xticks[0]
        max_x = xticks[-1]
        x_range = max_x - min_x
        if y_max_bar_plot is not None:
            axis.set_xlim((0, y_max_bar_plot))
        else:
            axis.set_xlim((min_x, max_x + 1))
        start, end = axis.get_xlim()
        if y_spacing_bar_plot is not None:
            axis.xaxis.set_ticks(np.arange(start, end + 1, y_spacing_bar_plot))

        if show_title:
            axis.set_title(plot_title)
        axis.set_ylabel("")
        if unit is None:
            if not english:
                x_label = "% van bedrijven met website"
            else:
                x_label = "% of companies with website"
        else:
            x_label = unit

        axis.set_xlabel(x_label, rotation="horizontal", horizontalalignment="right")
        axis.xaxis.set_label_coords(1.01, -0.12)
        axis.yaxis.grid(False)
        sns.despine(ax=axis, bottom=True)
        axis.tick_params(which="both", left=False)

        add_axis_label_background(
            fig=fig, axes=axis, loc="east", radius_corner_in_mm=1, margin=margin
        )

        number_of_columns = plot_df.columns.values.size
        if legend_position is None:
            legend_bbox_to_anchor = (0.02, 0.00)
        else:
            legend_bbox_to_anchor = legend_position
        legend_bbox_to_anchor = get_windows_or_linux_value(legend_bbox_to_anchor)
        axis.legend(
            loc="lower left",
            frameon=False,
            ncol=number_of_columns,
            bbox_to_anchor=legend_bbox_to_anchor,
            bbox_transform=fig.transFigure,
        )

    if reference_lines is not None:
        color = line_iter.get_next_color()
        for ref_key, ref_line in reference_lines.items():
            ref_label = ref_line["label"]
            ref_plot_df = ref_line["plot_df"]
            value = ref_plot_df.values[0][1]
            color = line_iter.get_next_color()
            axis.axhline(y=value, color=color, linestyle="-.")
            axis.text(
                xoff, value + yoff * x_range, ref_label, color=color, transform=trans
            )

    _logger.info(f"Saving plot {image_file_name}")
    fig.savefig(image_file)

    if highcharts_directory is not None:
        highcharts_directory.mkdir(exist_ok=True, parents=True)
    if export_svg:
        # met export highcharts gaan we ook een svg exporten
        svg_image_file = highcharts_directory / Path(
            "_".join([plot_key, image_name + ".svg"])
        )
        _logger.info(f"Saving plot {svg_image_file}")
        fig.savefig(svg_image_file)

    if export_highcharts:
        if y_max_bar_plot is not None:
            y_lim = (0, y_max_bar_plot)
        else:
            y_lim = None

        if title is not None:
            plot_title = title
        if barh:
            hc_ylabel = x_label
        else:
            hc_ylabel = y_label
        hc_file = "/".join([highcharts_directory.as_posix(), image_file.stem]) + ".json"
        _logger.info(f"Saving highcharts plot to: {hc_file}")
        CBSHighChart(
            data=plot_df,
            chart_type="bar_stacked",
            output_directory=highcharts_directory.as_posix(),
            output_file_name=image_file.stem,
            ylabel=hc_ylabel,
            y_lim=y_lim,
            y_tick_interval=y_spacing_bar_plot,
            title=plot_title,
            enable_legend=enable_highcharts_legend,
        )

    if show_plots:
        plt.show()

    plt.close()

    return image_file_name


def make_conditional_score_plot(
    correlations,
    image_directory,
    show_plots=False,
    figsize=None,
    image_type=".pdf",
    export_svg=False,
    export_highcharts=False,
    highcharts_directory=None,
    title=None,
    cache_directory=None,
    english=False,
):
    plot_info = correlations["plots"]

    index_labels = correlations["index_labels"]
    categories = correlations["index_categories"]
    score_intervallen = correlations["score_intervallen"]

    for plot_key, plot_prop in plot_info.items():

        # we maken hier alleen de score plots
        if plot_key not in (
            "scores_per_interval",
            "scores_per_number_correct",
        ) or not plot_prop.get("do_it", True):
            continue

        outfile = Path(plot_prop["output_file"])
        if cache_directory is not None:
            outfile = Path(cache_directory) / outfile
        in_file = outfile.with_suffix(".pkl")

        if highcharts_directory is None:
            hc_dir = Path(".")
        else:
            hc_dir = Path(highcharts_directory)

        if hc_sub_dir := plot_prop.get("highcharts_output_directory"):
            hc_dir = hc_dir / Path(hc_sub_dir)

        if hc_label := plot_prop.get("highcharts_label"):
            label = hc_label
        else:
            label = "Leeg"

        _logger.info(f"Reading scores from {in_file}")
        scores = pd.read_pickle(in_file.with_suffix(".pkl"))

        if plot_key == "scores_per_interval":
            x_label = AxisLabel(
                plot_prop.get("x_label"),
                text_default="Eindscoreniveau",
                positie_default=(0.98, -0.15),
            )
            y_label = AxisLabel(
                plot_prop.get("y_label"),
                text_default="Subgroepscore [%]",
                positie_default=(-0.065, 1.07),
            )

            im_file_base = "_".join([outfile.stem, "per_score_interval"])
            im_file = image_directory / Path(im_file_base).with_suffix(".pdf")
            plot_score_per_interval(
                scores=scores,
                score_intervallen=score_intervallen,
                index_labels=index_labels,
                categories=categories,
                highcharts_directory=hc_dir,
                im_file=im_file,
                show_plots=show_plots,
                plot_title=label,
                x_label=x_label,
                y_label=y_label,
                english=english,
            )
        elif plot_key == "scores_per_number_correct":
            x_label = AxisLabel(
                plot_prop.get("x_label"),
                text_default="Aantal geslaagde categorieën",
                positie_default=(0.98, -0.15),
            )
            y_label = AxisLabel(
                plot_prop.get("y_label"),
                text_default="Subgroepscore [%]",
                positie_default=(-0.065, 1.07),
            )

            im_file_base = "_".join([outfile.stem, "per_count_interval"])
            im_file = image_directory / Path(im_file_base).with_suffix(".pdf")
            plot_score_per_count(
                scores=scores,
                categories=categories,
                highcharts_directory=hc_dir,
                im_file=im_file,
                show_plots=show_plots,
                plot_title=label,
                x_label=x_label,
                y_label=y_label,
                english=english,
            )


def plot_score_per_count(
    scores,
    categories,
    highcharts_directory,
    im_file,
    show_plots,
    plot_title,
    x_label,
    y_label,
    english=False,
):
    _logger.info("Plot score per count")
    # add a new columns with the interval label belonging to the gk code bin. Note that we
    # merge all the grootte klass below 40 to a group smaller than 10

    score_per_category = dict()
    for categorie_key, category_df in scores.groupby("count"):
        _logger.debug(f"Plotting {categorie_key}")
        df = category_df[list(categories.keys())]
        score_per_category[categorie_key] = df.mean()

    score_per_category_df = pd.DataFrame(score_per_category).T * 100
    score_per_category_df = score_per_category_df.round(1)

    settings = CBSPlotSettings(color_palette="koelextended")
    fig, axis = plt.subplots()
    fig.subplots_adjust(bottom=0.3, top=0.90)
    score_per_category_df.plot.bar(
        ax=axis, rot=0, stacked=False, edgecolor="white", linewidth=1.5
    )
    yticks = axis.get_yticks()
    # axis.set_ylim((yticks[0], yticks[-1]))
    axis.set_ylim((0, 100))

    axis.set_xlabel(x_label.text, rotation="horizontal", horizontalalignment="right")
    axis.xaxis.set_label_coords(*x_label.positie)

    axis.set_ylabel(y_label.text, rotation="horizontal", horizontalalignment="left")
    axis.yaxis.set_label_coords(*y_label.positie)
    axis.xaxis.grid(False)
    sns.despine(ax=axis, left=True)

    # niet meer volgens de richtlijnen
    # add_values_to_bars(axis=axis, color="w")

    sns.despine(ax=axis, left=True)

    axis.tick_params(which="both", bottom=False)

    add_axis_label_background(fig=fig, axes=axis, loc="south")

    ncol = (score_per_category_df.columns.size - 1) // 2 + 1

    legend = axis.legend(
        loc="lower left",
        bbox_to_anchor=(0.105, -0.00),
        frameon=False,
        bbox_transform=fig.transFigure,
        ncol=ncol,
    )

    _logger.info(f"Writing score plot to {im_file}")
    fig.savefig(im_file.as_posix())

    highcharts_directory.mkdir(exist_ok=True, parents=True)

    CBSHighChart(
        data=score_per_category_df,
        chart_type="column_grouped",
        output_directory=highcharts_directory.as_posix(),
        output_file_name=im_file.stem,
        y_lim=(0, 100),
        ylabel=y_label.text,
        title=plot_title,
        enable_legend=True,
    )

    if show_plots:
        plt.show()

    _logger.debug("Klaar")


def plot_score_per_interval(
    scores,
    score_intervallen,
    index_labels,
    categories,
    highcharts_directory,
    im_file,
    show_plots,
    plot_title,
    x_label,
    y_label,
    english=False,
):
    score_labels = list(score_intervallen.keys())
    score_bins = list([s / 100 for s in score_intervallen.values()]) + [1.01]
    # add a new columns with the interval label belonging to the gk code bin. Note that we
    # merge all the grootte klass below 40 to a group smaller than 10
    scores["score_category"] = pd.cut(
        scores["score"],
        bins=score_bins,
        labels=score_labels,
        right=True,
        include_lowest=True,
    )

    score_per_category = dict()
    for categorie_key, category_df in scores.groupby("score_category"):
        _logger.debug(f"Plotting {categorie_key}")
        df = category_df[list(categories.keys())]
        category_label = index_labels[categorie_key]
        score_per_category[category_label] = df.mean()

    score_per_category_df = pd.DataFrame(score_per_category).T * 100
    score_per_category_df = score_per_category_df.round(1)

    settings = CBSPlotSettings(color_palette="koelextended")
    fig, axis = plt.subplots()
    fig.subplots_adjust(bottom=0.3, top=0.90)
    score_per_category_df.plot.bar(
        ax=axis, rot=0, stacked=False, edgecolor="white", linewidth=1.5
    )
    yticks = axis.get_yticks()
    # axis.set_ylim((yticks[0], yticks[-1]))
    axis.set_ylim((0, 100))

    axis.set_xlabel(x_label.text, rotation="horizontal", horizontalalignment="right")
    axis.xaxis.set_label_coords(*x_label.positie)

    axis.set_ylabel(y_label.text, rotation="horizontal", horizontalalignment="left")
    axis.yaxis.set_label_coords(*y_label.positie)
    axis.xaxis.grid(False)
    sns.despine(ax=axis, left=True)

    # niet meer volgens de richtlijnen
    # add_values_to_bars(axis=axis, color="w")

    sns.despine(ax=axis, left=True)

    axis.tick_params(which="both", bottom=False)

    add_axis_label_background(fig=fig, axes=axis, loc="south")

    ncol = (score_per_category_df.columns.size - 1) // 2 + 1

    legend = axis.legend(
        loc="lower left",
        bbox_to_anchor=(0.105, -0.00),
        frameon=False,
        bbox_transform=fig.transFigure,
        ncol=ncol,
    )

    _logger.info(f"Writing score plot to {im_file}")
    fig.savefig(im_file.as_posix())

    highcharts_directory.mkdir(exist_ok=True, parents=True)

    CBSHighChart(
        data=score_per_category_df,
        chart_type="column_grouped",
        output_directory=highcharts_directory.as_posix(),
        output_file_name=im_file.stem,
        y_lim=(0, 100),
        ylabel=y_label.text,
        title=plot_title,
        enable_legend=True,
    )

    if show_plots:
        plt.show()

    _logger.debug("Klaar")


# fig, axis = plt.subplots(figsize=(10, 10))
# cbar_ax = fig.add_axes([.91, .315, .02, .62])
# cmap = sns.color_palette("deep", 10)


def make_heatmap(
    correlations,
    image_directory,
    show_plots=False,
    figsize=None,
    image_type=".pdf",
    export_svg=False,
    export_highcharts=False,
    highcharts_directory=None,
    title=None,
    cache_directory=None,
    english=False,
):
    plot_properties = correlations["plots"]["correlation"]
    outfile = Path(plot_properties["output_file"])
    if cache_directory is not None:
        outfile = Path(cache_directory) / outfile

    in_file = outfile.with_suffix(".pkl")

    if highcharts_directory is None:
        hc_dir = Path(".")
    else:
        hc_dir = Path(highcharts_directory)

    if hc_sub_dir := plot_properties.get("highcharts_output_directory"):
        hc_dir = highcharts_directory / Path(hc_sub_dir)

    _logger.info(f"Reading correlation from {in_file}")
    corr = pd.read_pickle(in_file.with_suffix(".pkl"))

    categories = correlations["index_categories"]
    corr_index = correlations["index_correlations"]
    corr = corr.reindex(list(corr_index.keys()))
    corr = corr[list(corr_index.keys())]

    sns.set(font_scale=0.8)
    # cmap is now a list of colors
    cmap = mpc.ListedColormap(
        sns.cubehelix_palette(start=2.8, rot=0.1, light=0.9, n_colors=12)
    )

    # Create two appropriately sized subplots
    # grid_kws = {'width_ratios': (0.9, 0.03), 'wspace': 0.18}
    # fig, (axis, cbar_ax) = plt.subplots(1, 2, gridspec_kw=grid_kws, figsize=(8.3, 8.3))

    im_file = image_directory / Path(outfile.stem).with_suffix(".pdf")
    fig, axis = plt.subplots(figsize=(10, 10))
    plt.subplots_adjust(left=0.28, bottom=0.27, top=0.98, right=0.9)
    cbar_ax = fig.add_axes([0.91, 0.315, 0.02, 0.62])
    # cmap = sns.color_palette("deep", 10)

    sns.heatmap(
        corr,
        square=True,
        ax=axis,
        cbar_ax=cbar_ax,
        cmap=cmap,
        vmin=-0.2,
        vmax=1.0,
        cbar_kws={"orientation": "vertical", "label": r"Correlatiecoëfficiënt $\rho$"},
    )
    xlabels = axis.get_xticklabels()
    ylabels = axis.get_yticklabels()
    for xlbl, ylbl in zip(xlabels, ylabels):
        tekst = xlbl.get_text()
        categorie = corr_index[tekst]
        categorie_properties = categories[categorie]
        kleur = categorie_properties["color"]
        RGB = CBS_COLORS_RBG.get(kleur, [0, 0, 0])
        rgb = [_ / 255 for _ in RGB]
        tekst_clean = tekst.replace("_verdict", "").replace("tests_", "")
        xlbl.set_text(tekst_clean)
        xlbl.set_color(rgb)
        ylbl.set_text(tekst_clean)
        ylbl.set_color(rgb)

    axis.set_xticklabels(xlabels, rotation=90, ha="right")
    axis.set_yticklabels(ylabels, rotation=0, ha="right")

    plt.legend(loc="upper left", prop={"size": 10})

    _logger.info(f"Writing heatmap to {im_file}")
    fig.savefig(im_file.as_posix())

    hc_dir.mkdir(exist_ok=True, parents=True)

    hc_out = hc_dir / Path(im_file.stem + ".svg")

    _logger.info(f"Writing heatmap to {hc_out}")
    fig.savefig(hc_out.as_posix())

    if show_plots:
        plt.show()


def make_conditional_pdf_plot(
    categories,
    image_directory,
    show_plots=False,
    export_highcharts=False,
    highcharts_directory=None,
    cache_directory=None,
    english=False,
):
    outfile = Path(categories["categories_output_file"])
    if cache_directory is not None:
        outfile = Path(cache_directory) / outfile

    image_key = "pdf_per_category"
    plot_settings = categories["plot_settings"]["pdf_per_category"]
    y_max = plot_settings.get("y_max_pdf_plot")
    y_spacing = plot_settings.get("y_spacing_pdf_plot")
    export_svg = plot_settings.get("export_svg")

    in_file = outfile.with_suffix(".pkl")

    if highcharts_directory is None:
        hc_dir = Path(".")
    else:
        hc_dir = Path(highcharts_directory)

    if hc_sub_dir := plot_settings.get("highcharts_output_directory"):
        hc_dir = hc_dir / Path(hc_sub_dir)

    hc_dir.mkdir(exist_ok=True, parents=True)

    _logger.info(f"Reading correlation from {in_file}")
    conditional_scores_df = pd.read_pickle(in_file.with_suffix(".pkl"))

    im_file = image_directory / Path(outfile.stem).with_suffix(".pdf")

    figure_properties = CBSPlotSettings()

    fig, axis = plt.subplots()
    axis.tick_params(which="both", bottom=True)
    delta_bin = np.diff(conditional_scores_df.index)[0]

    fig.subplots_adjust(bottom=0.25, top=0.92, right=0.98)
    axis.tick_params(which="both", bottom=True)

    conditional_scores_df.index = conditional_scores_df.index + delta_bin / 2

    for col_name in conditional_scores_df.columns:
        pdf = 100 * conditional_scores_df[col_name].to_numpy()
        axis.bar(conditional_scores_df.index, pdf, width=delta_bin, label=col_name)

    xtics = np.linspace(0, 100, endpoint=True, num=6)
    _logger.debug(xtics)
    _logger.debug(conditional_scores_df.index)
    axis.xaxis.set_ticks(xtics)
    axis.set_xlim((-5, 105))

    start, end = axis.get_ylim()
    if y_max is not None:
        end = y_max
    if y_spacing is not None:
        axis.yaxis.set_ticks(np.arange(start, end + 1, y_spacing))

    if y_max is not None:
        axis.set_ylim((0, y_max))

    # this triggers the drawing, otherwise we can not retrieve the xtick labels
    fig.canvas.draw()

    #    y_label = '% van bedrijven met website'
    y_label = "% of companies with website"

    axis.set_ylabel(y_label, rotation="horizontal", horizontalalignment="left")
    axis.yaxis.set_label_coords(-0.04, 1.05)
    axis.xaxis.grid(False)
    axis.set_xlabel("Eindscore", horizontalalignment="right")
    axis.xaxis.set_label_coords(0.98, -0.12)
    sns.despine(ax=axis, left=True)

    labels = [_.get_text() for _ in axis.get_xticklabels()]
    axis.set_xticklabels(labels, ha="center")

    add_axis_label_background(fig=fig, axes=axis, loc="south", margin=0.10)

    if not english:
        plot_title = "Aantal geslaagde categorieën"
        hc_plot_title = "Verdeling scores per categorie"
    else:
        plot_title = "Number of succeeded categories "
        hc_plot_title = "Distribution of scores per category"

    legend = axis.legend(
        loc="lower left",
        title=plot_title,
        prop={"size": 10},
        bbox_to_anchor=(0.2, 0.02),
        frameon=False,
        bbox_transform=fig.transFigure,
        ncol=5,
    )

    legend._legend_box.align = "left"
    for patch in legend.get_patches():
        patch.set_linewidth(0)

    _logger.info(f"Saving plot {im_file}")
    fig.savefig(im_file)

    if export_svg:
        svg_image_file = hc_dir / Path(im_file.stem).with_suffix(".svg")
        _logger.info(f"Saving plot to {svg_image_file}")
        svg_image_file.parent.mkdir(exist_ok=True, parents=True)
        fig.savefig(svg_image_file.as_posix())

    if export_highcharts:
        # voor highcharts de titel setten
        CBSHighChart(
            data=conditional_scores_df,
            chart_type="column",
            output_directory=hc_dir.as_posix(),
            output_file_name=im_file.stem,
            ylabel=y_label,
            title=hc_plot_title,
            enable_legend=False,
        )

    if show_plots:
        plt.show()

    _logger.debug("Done")

    plt.close()


def make_verdeling_per_aantal_categorie(
    categories,
    image_directory,
    show_plots=False,
    export_highcharts=False,
    highcharts_directory=None,
    cache_directory=None,
    english=False,
):
    outfile = Path(categories["categories_output_file"])
    if cache_directory is not None:
        outfile = Path(cache_directory) / outfile

    image_key = "verdeling_per_categorie"
    plot_settings = categories["plot_settings"][image_key]
    y_max = plot_settings.get("y_max_pdf_plot")
    y_spacing = plot_settings.get("y_spacing_pdf_plot")
    export_svg = plot_settings.get("export_svg")

    index_categories = categories["index_categories"]
    renames = dict()
    for index_key, index_prop in index_categories.items():
        variable_name = index_prop["variable"]
        renames[variable_name] = index_key

    in_file = outfile.with_suffix(".pkl")
    sum_file = in_file.parent / Path(in_file.stem + "_sum.pkl")
    _logger.info(f"Reading from {sum_file}")
    sum_per_number_of_cat_df = pd.read_pickle(sum_file)
    sum_per_number_of_cat_df.rename(columns=renames, inplace=True)
    # zet de volgorde gelijk aan de settings file
    sum_per_number_of_cat_df = sum_per_number_of_cat_df[list(index_categories.keys())]

    sum_per_number_of_cat_df = sum_per_number_of_cat_df.T
    sum_per_number_of_cat_df.drop(0, axis=1, inplace=True)

    sum_of_all_categories = sum_per_number_of_cat_df.sum()

    percentage_per_number_of_cat = (
        100 * sum_per_number_of_cat_df / sum_of_all_categories
    )

    x_label = "undefined x"
    y_label = "undefined y"
    if hc_title := plot_settings.get("highcharts_label"):
        title = hc_title
        hc_plot_title = "undefined"
        l_label = "undefined"
    else:
        if not english:
            x_label = "Aantal geslaagde categorieën"
            y_label = "Aandeel per categorie [%]"
            l_label = "Categorie"
            hc_plot_title = "Verdeling scores per categorie"
        else:
            x_label = "Number of succeeded categories"
            y_label = "Part per category [%]"
            l_label = "Category"
            hc_plot_title = "Distribution of scores per category"

    if highcharts_directory is None:
        hc_dir = Path(".")
    else:
        hc_dir = Path(highcharts_directory)

    if hc_sub_dir := plot_settings.get("highcharts_output_directory"):
        hc_dir = hc_dir / Path(hc_sub_dir)

    hc_dir.mkdir(exist_ok=True, parents=True)

    im_file = image_directory / Path("_".join([outfile.stem, image_key])).with_suffix(
        ".pdf"
    )

    figure_properties = CBSPlotSettings()

    fig, axis = plt.subplots()
    axis.tick_params(which="both", bottom=True)
    fig.subplots_adjust(bottom=0.25, top=0.92, right=0.98)

    percentage_per_number_of_cat.T.plot.bar(stacked=True, ax=axis)

    axis.set_ylim((0, 101))

    axis.set_xlabel(x_label, horizontalalignment="right")

    axis.set_ylabel(y_label, rotation="horizontal", horizontalalignment="left")
    axis.yaxis.set_label_coords(-0.06, 1.05)
    axis.xaxis.grid(False)
    axis.xaxis.set_label_coords(0.98, -0.1)
    xlabels = axis.get_xticklabels()
    axis.set_xticklabels(xlabels, rotation=0, ha="right")
    sns.despine(ax=axis, left=True)

    legend = axis.legend(
        loc="lower left",
        title=l_label,
        bbox_to_anchor=(0.2, 0.03),
        frameon=False,
        bbox_transform=fig.transFigure,
        ncol=5,
    )

    legend._legend_box.align = "left"
    for patch in legend.get_patches():
        patch.set_linewidth(0)

    axis.tick_params(which="both", bottom=False)
    add_axis_label_background(fig=fig, axes=axis, loc="south", margin=0.02)

    _logger.info(f"Saving plot {im_file}")
    fig.savefig(im_file)

    if export_svg:
        svg_image_file = hc_dir / Path(im_file.stem).with_suffix(".svg")
        _logger.info(f"Saving plot to {svg_image_file}")
        fig.savefig(svg_image_file)

    if export_highcharts:
        # voor highcharts de titel setten
        CBSHighChart(
            data=percentage_per_number_of_cat.T,
            chart_type="column_stacked_percentage",
            output_directory=hc_dir.as_posix(),
            output_file_name=im_file.stem,
            y_lim=(0, 100),
            title=hc_plot_title,
            xlabel=x_label,
            ylabel=y_label,
            enable_legend=True,
        )

    if show_plots:
        plt.show()

    _logger.debug("Done")

    plt.close()



from internetnl_domain_analyse.utils import prepare_stat_data_for_write

__author__ = "Eelco van Vliet"
__copyright__ = "Eelco van Vliet"
__license__ = "MIT"


def test_prepare_data():

    assert prepare_stat_data_for_write is not None



import codecs
import logging
import os
import pickle
import re
import sqlite3
import sys
from collections import Counter
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas import DataFrame, Series
import yaml

from weighted_sample_statistics import WeightedSampleStatistics, VariableProperties
from weighted_sample_statistics import (
    get_records_select,
    rename_all_variables,
    prepare_df_for_statistics,
)
from internetnl_domain_analyse.domain_plots import (
    make_cdf_plot,
    make_bar_plot,
    make_bar_plot_stacked,
)
from internetnl_domain_analyse.latex_output import make_latex_overview
from internetnl_domain_analyse.utils import (
    read_tables_from_sqlite,
    get_all_clean_urls,
    dump_data_frame_as_sqlite,
    add_derived_variables,
    fill_booleans,
    prepare_stat_data_for_write,
    get_option_mask,
    impose_variable_defaults,
    add_missing_groups,
    clean_all_suffix,
    get_windows_or_linux_value,
)

_logger = logging.getLogger(__name__)

tld_logger = logging.getLogger("tldextract")

mpl_logger = logging.getLogger("matplotlib")
mpl_logger.setLevel(logging.WARNING)


def make_plot_cache_file_name(cache_directory, file_base, prefix):
    return cache_directory / Path("_".join([prefix, file_base, "cache_for_plot.pkl"]))


class ImageFileInfo:
    def __init__(
        self, scan_data_key, cache_file_name_base="image_info", cache_directory="cache"
    ):
        self.scan_data_key = scan_data_key
        self.cache_directory = Path(cache_directory)
        self.cache_directory.mkdir(exist_ok=True)
        cache_file_name = Path(
            "_".join([cache_file_name_base, scan_data_key])
        ).with_suffix(".yml")
        self.cache_file_name = self.cache_directory / cache_file_name

        self.data = None

    def add_entry(
        self,
        plot_key,
        plot_info,
        image_key,
        sub_image_label,
        file_name,
        tex_right_shift=None,
        section=None,
    ):
        """add a new entry"""

        if image_key not in self.data.keys():
            self.data[image_key] = dict()
        self.data[image_key][plot_key] = dict(
            file_name=file_name,
            tex_right_shift=tex_right_shift,
            sub_image_label=sub_image_label,
        )
        if section:
            self.data[image_key][plot_key]["section"] = section

        # To get the key order in the dict the same as in the input file, alter the order
        if len(self.data[image_key].keys()) > 1:
            tmp_data = self.data[image_key].copy()
            self.data[image_key] = dict()
            for plot_key in plot_info.keys():
                try:
                    tex_prop = tmp_data[plot_key]
                except KeyError as err:
                    # this entry is not in the plot. No problem. skip it
                    pass
                else:
                    self.data[image_key][plot_key] = tex_prop

    def fix_order(self, variables):
        tmp_info = self.data.copy()
        self.data = dict()
        for var_name in variables.index:
            try:
                var_entry = tmp_info[var_name]
            except KeyError as err:
                _logger.debug(f"no entry for {var_name}. No problem, skipping")
            else:
                self.data[var_name] = var_entry

    def read_cache(self):
        """Lees de cache"""
        if self.cache_file_name.exists():
            with codecs.open(
                self.cache_file_name.as_posix(), "r", encoding="UTF-8"
            ) as stream:
                self.data = yaml.load(stream=stream, Loader=yaml.Loader)
        else:
            self.data = dict()

    def write_cache(self):
        """Schrijf de cache"""
        with codecs.open(
            self.cache_file_name.as_posix(), "w", encoding="UTF-8"
        ) as stream:
            yaml.dump(data=self.data, stream=stream, Dumper=yaml.Dumper)


class RecordCacheInfo:
    def __init__(
        self, records_cache_data: dict, year_key: str, stat_directory: str = None
    ):
        """
        Store the properties of the cache file in a class
        """
        self.records_cache_data = records_cache_data
        self.stat_directory = stat_directory

        self.year_key = f"{year_key}"
        match = re.search("20(\d\d)", self.year_key)
        if match:
            self.year_digits = match.group(1)
        else:
            self.year_digits = year_key[-2:]

        self.cache_dir = None
        self.file_name = None
        self.table_names = None

        self.get_cache_file_name()
        self.get_cache_table_names()

    def get_cache_file_name(self):
        """
        Retrieve the cache file name from the dictionary. If environment variables are given,
        base the directory on the environment name. Names are given like RECORDS_CACHE_DIR_20,
        RECORDS_CACHE_DIR_21, for 2020, 2021 resp.
        """

        records_environment_variable = "_".join(["RECORDS_CACHE_DIR", self.year_key])
        records_cache_dir_name = os.getenv(records_environment_variable)

        if records_cache_dir_name is None:
            records_cache_dir_name = self.records_cache_data.get(
                "records_cache_directory", "."
            )

        records_cache_dir_name = records_cache_dir_name.replace(
            "{{ stat_directory }}", self.stat_directory
        )

        self.cache_dir = Path(records_cache_dir_name)

        records_file_basename = Path(self.records_cache_data["records_cache_file"])

        self.file_name = self.cache_dir / records_file_basename

    def get_cache_table_names(self):
        """
        Get the table names of the cache files.
        """

        table_records_environment_variable = "_".join(
            ["RECORDS_TABLE_RECS", self.year_key]
        )
        table_records_name = os.environ.get(table_records_environment_variable)
        table_info_environment_variable = "_".join(
            ["RECORDS_TABLE_INFO", self.year_key]
        )
        tabl_info_name = os.environ.get(table_info_environment_variable)

        if table_records_name is None:
            # the environment variabel RECORDS_TABLE_INFO is not set. read from the settings file
            self.table_names = self.records_cache_data.get("records_table_names")
            if self.table_names is None:
                # It is also not found in the settings file. Make a guess
                self.table_names = [
                    f"records_df_{self.year_digits}_2",
                    f"info_records_df_{self.year_digits}",
                ]
        else:
            self.table_names = [table_records_name, tabl_info_name]


class DomainAnalyser:
    def __init__(
        self,
        scan_data_key=None,
        cache_file_base="tables_df",
        cache_directory_base_name=None,
        tld_extract_cache_directory=None,
        output_file=None,
        reset=None,
        records_cache_info: RecordCacheInfo = None,
        internet_nl_filename=None,
        breakdown_labels=None,
        statistics: dict = None,
        default_scan=None,
        variables: dict = None,
        module_info: dict = None,
        weights=None,
        url_key="website_url",
        suffix_key="suffix",
        translations=None,
        module_key="module",
        variable_key="variable",
        sheet_renames=None,
        n_digits=None,
        write_dataframe_to_sqlite=False,
        statistics_to_xls=False,
        n_bins=100,
        mode=None,
        correlations=None,
        categories=None,
        dump_cache_as_sqlite=False,
    ):

        _logger.info(f"Running here {os.getcwd()}")

        self.records_cache_info = records_cache_info

        if output_file is None:
            self.output_file = Path("output.sqlite")
        else:
            self.output_file = Path(output_file)
        self.output_directory = Path("output")
        self.output_directory.mkdir(exist_ok=True)
        self.output_file = self.output_directory / self.output_file

        outfile_suff = self.output_file.suffixes
        outfile_base = self.output_file.with_suffix("").with_suffix("").as_posix()
        outfile_year = Path(
            "_".join([outfile_base, scan_data_key, self.records_cache_info.year_key])
        )
        self.output_file = outfile_year.with_suffix(".".join(outfile_suff))
        self.dump_cache_as_sqlite = dump_cache_as_sqlite

        self.scan_data_key = scan_data_key
        self.breakdown_labels = breakdown_labels

        self.correlations = correlations
        self.categories = categories
        self.statistics = statistics
        self.default_scan = default_scan
        self.module_key = module_key
        self.variable_key = variable_key
        self.module_info = module_info
        self.variables = self.variable_dict2df(variables, module_info)
        self.n_digits = n_digits
        self.n_bins = n_bins

        self.sheet_renames = sheet_renames

        self.url_key = url_key
        self.suffix_key = suffix_key
        self.be_id = "be_id"
        self.mi_labels = ["sbi", "gk_sbs", self.be_id]
        self.translations = translations

        self.categories_coefficient_df = None
        self.correlation_coefficient_df = None

        if internet_nl_filename is not None:
            self.internet_nl_filename = internet_nl_filename
        else:
            self.internet_nl_filename = Path("internet_nl.sqlite")

        self.cache_directory = Path(
            "_".join([cache_directory_base_name, self.records_cache_info.year_key])
        )
        self.cache_directory.mkdir(exist_ok=True)
        if tld_extract_cache_directory is None:
            self.tld_extract_cache_directory = "tld_cache"
        else:
            self.tld_extract_cache_directory = tld_extract_cache_directory
        cache_file_base = Path(
            "_".join([cache_file_base, self.records_cache_info.year_key, scan_data_key])
            + ".pkl"
        )
        self.cache_file = self.cache_directory / cache_file_base
        self.cate_outfile = None
        self.cate_pkl_file = None
        self.corr_outfile = None
        self.corr_pkl_file = None
        self.score_outfile = None
        self.score_pkl_file = None
        if self.correlations is not None:
            plot_info = self.correlations["plots"]
        else:
            plot_info = None
        try:
            self.cate_outfile = self.cache_directory / Path(
                self.categories["categories_output_file"]
            )
        except TypeError:
            _logger.debug("categories not defined")
        else:
            self.cate_pkl_file = self.cate_outfile.with_suffix(".pkl")
        try:
            self.corr_outfile = self.cache_directory / Path(
                plot_info["correlation"]["output_file"]
            )
        except TypeError:
            _logger.debug("correlations not defined")
        else:
            self.corr_pkl_file = self.corr_outfile.with_suffix(".pkl")
            self.score_outfile = self.cache_directory / Path(
                plot_info["scores_per_interval"]["output_file"]
            )
            self.score_pkl_file = self.score_outfile.with_suffix(".pkl")

        if reset is None:
            self.reset = None
        else:
            self.reset = int(reset)
        self.weight_key = weights

        self.dataframe = None
        self.score_df = None
        self.all_stats_per_format = dict()
        self.all_hist_per_format = dict()

        self.all_plots = None

        have_cache = self.check_if_cache_exist(mode)

        if (self.reset is not None and self.reset <= 1) or not have_cache:
            # de microdata alleen lezen als we geen pickle files van de statistische output hebben
            # als we alleen plaatjes willen maken is het sneller om de uitgerekende tabellen van
            # cache te lezen
            self.read_data()

        if write_dataframe_to_sqlite:
            if self.dataframe is not None:
                self.write_data()
                sys.exit(0)
            else:
                msg = (
                    "The write_data_frame option only works if you force to read the micro "
                    "data with --reset 0 or --reset 1"
                )
                _logger.warning(msg)

        if mode in ("all", "statistics"):
            self.calculate_statistics()
            if statistics_to_xls or reset is not None or not have_cache:
                self.write_statistics()
        if mode in ("all", "correlations") and self.dataframe is not None:
            self.calculate_correlations_and_scores()
        if mode in ("all", "categories") and self.dataframe is not None:
            self.calculate_categories()

    def check_if_cache_exist(self, mode: str):

        cache_exists = True
        if mode in ("all", "statistics"):
            cache_exists = self.cache_file.exists()
        if mode in ("all", "correlations"):
            cache_exists = cache_exists and self.corr_pkl_file.exists()
            cache_exists = cache_exists and self.score_pkl_file.exists()
        if mode in ("all", "categories"):
            cache_exists = cache_exists and self.cate_pkl_file.exists()

        return cache_exists

    def variable_dict2df(self, variables, module_info: dict = None) -> DataFrame:
        """
        Converteer de directory met variable info naar een data frame
        Args:
            variables:  dict met variable info
            module_info: dict met module informatie

        Returns:
            dataframe
        """
        var_df = pd.DataFrame.from_dict(variables).unstack().dropna()
        var_df = var_df.reset_index()
        var_df = var_df.rename(
            columns={
                "level_0": self.module_key,
                "level_1": self.variable_key,
                0: "properties",
            }
        )
        var_df.set_index(self.variable_key, drop=True, inplace=True)

        var_df = impose_variable_defaults(
            var_df, module_info=module_info, module_key=self.module_key
        )
        return var_df

    def write_statistics(self):
        _logger.info(f"Writing statistics {self.output_file}")
        connection = sqlite3.connect(self.output_file)

        excel_file = Path(self.output_file).with_suffix(".xlsx")
        sheets = list()
        cnt = 0
        with pd.ExcelWriter(str(excel_file), engine="openpyxl") as writer:
            _logger.info(f"Start writing standard output to {excel_file}")

            for file_base, all_stats in self.all_stats_per_format.items():

                stat_df = prepare_stat_data_for_write(
                    file_base=file_base,
                    all_stats=all_stats,
                    variables=self.variables,
                    variable_key=self.variable_key,
                    module_key=self.module_key,
                    breakdown_labels=self.breakdown_labels,
                    n_digits=self.n_digits,
                    connection=connection,
                )

                cache_file = make_plot_cache_file_name(
                    cache_directory=self.cache_directory,
                    prefix=self.scan_data_key,
                    file_base=file_base,
                )
                _logger.info(f"Writing cache for stat {cache_file}")
                with open(cache_file, "wb") as stream:
                    pickle.dump(stat_df, stream)

                sheet_name = file_base
                if self.sheet_renames is not None:
                    for rename_key, sheet_rename in self.sheet_renames.items():
                        pat = sheet_rename["pattern"]
                        rep = sheet_rename["replace"]
                        sheet_name = re.sub(pat, rep, sheet_name)
                if len(sheet_name) > 32:
                    sheet_name = sheet_name[:32]
                if sheet_name in sheets:
                    sheet_name = sheet_name[:30] + "{:02d}".format(cnt)
                cnt += 1
                sheets.append(sheets)
                stat_df.to_excel(excel_writer=writer, sheet_name=sheet_name)

    def calculate_statistics_one_breakdown(self, group_by):

        index_names = group_by + [self.be_id]
        try:
            dataframe = prepare_df_for_statistics(
                self.dataframe, index_names=index_names, units_key="units"
            )
        except KeyError:
            _logger.info(f"Breakdown on {index_names} does not exist")
            return None, None

        all_stats = dict()
        all_hist = dict()

        for var_key, var_prop in self.variables.iterrows():
            _logger.debug(f"{var_key}")
            var_prop_klass = VariableProperties(
                variables=self.variables, column=var_key
            )

            column = var_key
            column_list = list([var_key])
            var_module = var_prop["module"]
            try:
                module = self.module_info[var_module]
            except KeyError as err:
                _logger.warning(err)
                continue
            if not module.get("include", True):
                continue

            var_type = var_prop["type"]
            var_filter = var_prop["filter"]
            var_weight_key = var_prop["gewicht"]
            schaal_factor_key = "_".join(["ratio", var_weight_key])
            units_schaal_factor_key = "_".join(["ratio", "units"])
            weight_cols = set(
                list([var_weight_key, schaal_factor_key, units_schaal_factor_key])
            )
            df_weights = dataframe.loc[:, list(weight_cols)]

            try:
                data, column_list = get_records_select(
                    dataframe=dataframe,
                    variables=self.variables,
                    var_type=var_type,
                    column=column,
                    column_list=column_list,
                    output_format="statline",
                    var_filter=var_filter,
                )
            except KeyError:
                _logger.info(f"Failed to get selection of {column}. Skipping")
                continue

            if data is None:
                _logger.info(f"Could not get data selection for {var_key}. Skipping")
                continue

            if data.index.size < df_weights.index.size:
                _logger.info(
                    f"we filtered data, reducing from {df_weights.index.size} to {data.index.size}"
                )
                df_weights = df_weights.reindex(data.index)

            stats = WeightedSampleStatistics(
                group_keys=group_by,
                records_df_selection=data,
                weights_df=df_weights,
                column_list=column_list,
                var_type=var_type,
                var_weight_key=var_weight_key,
                scaling_factor_key=schaal_factor_key,
                units_scaling_factor_key=units_schaal_factor_key,
                report_numbers=var_prop_klass.report_number,
            )

            stats.calculate()

            if (
                not np.isnan(var_prop_klass.report_number)
                and var_prop_klass.report_number
            ):
                all_stats[column] = stats.records_sum
            else:
                all_stats[column] = stats.records_weighted_mean_agg

            # voeg hier het histogram van de data toe
            all_hist[var_key] = calculate_histogram_per_breakdown(
                data, var_key=var_key, df_weights=df_weights, n_bins=self.n_bins
            )
        return all_stats, all_hist

    def get_correct_categories_count(self):
        """Bekijk per record hoeveel categorieën goed zijn en geef terug als dataframe"""

        col_sel = list()

        for cat_key, cat_prop in self.categories["index_categories"].items():
            variable = cat_prop["variable"]
            col_sel.append(variable)

        _logger.debug(f"make selection\n{col_sel}")
        data_df: pd.DataFrame = self.dataframe[col_sel]

        # alleen 1 wordt als succes beschouwd
        data_df = data_df == 1

        count = data_df.sum(axis=1)
        count = count.rename("count")

        return data_df, count

    def calculate_categories(self):
        if self.cate_pkl_file.exists() and self.reset is None:
            _logger.info(
                f"Cache {self.cate_pkl_file} and already exist. "
                f"Skip calculation categories and go to plot"
            )
            return
        if self.dataframe is None:
            msg = "For correlations you need the microdata. Run with --reset 1"
            raise ValueError(msg)

        _logger.info("Calculating categories")

        score_df = self.dataframe["percentage"].copy()
        score_df = score_df.rename("score")
        weights = self.dataframe[self.weight_key].copy()
        data_df, count = self.get_correct_categories_count()

        tot = pd.concat([score_df, count], axis=1)

        conditional_scores = list()
        sum_per_number_of_cat = list()

        total_sum = 0
        mask_tot: pd.Series = None
        for number_of_cat in range(0, 5):
            mask = tot["count"] == number_of_cat
            tot_cond = tot.loc[mask, "score"]
            sel_df = data_df[mask]
            sum_per_number_of_cat.append(sel_df.sum(axis=0))
            ww = weights[mask].to_numpy()
            if mask_tot is None:
                mask_tot = mask
            else:
                mask_tot = mask_tot | mask
            hist, bin_edge = np.histogram(
                tot_cond.to_numpy(),
                weights=ww,
                density=False,
                range=(0, 100),
                bins=self.n_bins,
            )
            hist_sum = hist.sum()
            total_sum += hist_sum
            conditional_scores.append(hist)

        sum_per_number_of_cat_df = pd.DataFrame.from_records(sum_per_number_of_cat)
        bin_width = bin_edge[1] - bin_edge[0]

        conditional_scores_df = pd.DataFrame().from_records(conditional_scores)
        conditional_scores_df.index = conditional_scores_df.index.rename("n_categories")
        conditional_scores_df = conditional_scores_df.T
        conditional_scores_df.index = bin_edge[:-1]
        conditional_scores_df /= total_sum * bin_width

        check_sum = conditional_scores_df.sum().sum() * bin_width

        _logger.debug(f"sum {check_sum}")

        _logger.info(f"Writing to {self.cate_pkl_file}")
        conditional_scores_df.to_pickle(self.cate_pkl_file)

        sum_file = self.cate_pkl_file.parent / Path(
            self.cate_pkl_file.stem + "_sum.pkl"
        )
        _logger.info(f"Writing to {sum_file}")
        sum_per_number_of_cat_df.to_pickle(sum_file)

    def calculate_correlations_and_scores(self):

        if (
            self.corr_pkl_file.exists()
            and self.score_pkl_file.exists()
            and self.reset is None
        ):
            _logger.info(
                f"Cache {self.corr_pkl_file} and {self.score_pkl_file} already exist. "
                f"Skip calculation and go to plot"
            )
            return

        if self.dataframe is None:
            msg = "For correlations you need the microdata. Run with --reset 1"
            raise ValueError(msg)

        data_df_count, count = self.get_correct_categories_count()

        index_columns = self.correlations["index_correlations"]

        _logger.info("Calculating correlations")
        col_sel = list(index_columns.keys())

        _logger.debug(f"make selection\n{col_sel}")
        data_df: pd.DataFrame = self.dataframe[col_sel]

        # alleen 1 wordt als succes beschouwd
        data_df = data_df == 1

        # verkrijg de categorieën van variabele met hoge correlatie
        categories = dict()
        for col_name, categorie in index_columns.items():
            try:
                categories[categorie].append(col_name)
            except KeyError:
                categories[categorie] = [col_name]

        # bereken de score per category en vergelijk met de internet.nl-score
        self.score_df = self.dataframe[["percentage"]].copy() / 100
        self.score_df.rename(columns={"percentage": "score"}, inplace=True)
        for categorie, columns in categories.items():
            selection = data_df[columns]
            max_score = len(columns)
            self.score_df[categorie] = selection.sum(axis=1) / max_score

        self.score_df = pd.concat([self.score_df, count], axis=1)

        desc = data_df.describe()
        _logger.debug(f"making descr\n{desc}")
        # reken correlatie twee keer uit
        corr = data_df.corr()
        ordered_index = corr.sum().sort_values(ascending=False).index
        data_df = data_df[ordered_index]
        corr = data_df.corr()
        self.correlation_coefficient_df = corr

        _logger.info(f"Schrijf naar {self.corr_outfile}")
        with sqlite3.connect(str(self.corr_outfile)) as connection:
            corr.to_sql(name="correlations", con=connection, if_exists="replace")
        _logger.info(f"Schrijf naar {self.corr_pkl_file}")
        corr.to_pickle(self.corr_pkl_file.as_posix())

        _logger.info(f"Schrijf naar {self.score_pkl_file}")
        self.score_df.to_pickle(self.score_pkl_file.as_posix())
        _logger.debug(f"making corrected\n{corr}")

    def calculate_statistics(self):
        _logger.info("Calculating statistics")

        self.all_stats_per_format = dict()
        self.all_hist_per_format = dict()
        missing_groups = None

        for file_base, props in self.statistics.items():
            scan_data = props.get("scan_data", self.scan_data_key)
            if scan_data != self.scan_data_key:
                _logger.debug(f"SKipping {scan_data} for {self.scan_data_key}")
                continue

            if not props.get("do_it", True):
                _logger.debug(
                    f"SKipping breakdown {file_base} for {self.scan_data_key}"
                )
                continue

            _logger.info(f"Processing {file_base}")

            file_name = Path("_".join([file_base, self.scan_data_key]) + ".pkl")
            cache_file = self.cache_directory / file_name

            group_by = list(props["groupby"].values())
            group_by_original = None
            if (
                group_by_if_not_exist := props.get("groupby_if_not_exist")
            ) and self.dataframe is not None:
                have_missing_groups = False
                for group in group_by:
                    if group not in self.dataframe.columns:
                        have_missing_groups = True
                if have_missing_groups:
                    group_by_original = group_by
                    group_by = list(group_by_if_not_exist.values())
                    missing_groups = props.get("missing_groups")

            combination: list = props.get("combination")

            if combination is None:
                if cache_file.exists() and self.reset is None:
                    _logger.info(f"Reading stats from cache {cache_file}")
                    with open(str(cache_file), "rb") as stream:
                        stat_df, all_hist = pickle.load(stream)
                elif self.dataframe is not None:
                    _logger.info("Calculating statistics from micro data")
                    all_stats, all_hist = self.calculate_statistics_one_breakdown(
                        group_by=group_by
                    )
                    if group_by_original is not None:
                        all_stats = add_missing_groups(
                            all_stats, group_by, group_by_original, missing_groups
                        )

                    if all_stats is None:
                        _logger.info(
                            f"Could not calculate statistics for breakdown {group_by}. Skip"
                        )
                        continue

                    # maak er een pandas data frame van
                    stat_df = pd.concat(list(all_stats.values()), axis=1, sort=False)
                    is_nan = stat_df.index == "nan"
                    stat_df = stat_df.loc[~is_nan]

                    _logger.info(f"Writing stats to cache {cache_file}")
                    with open(str(cache_file), "wb") as stream:
                        pickle.dump([stat_df, all_hist], stream)
                else:
                    _logger.info(f"Statistics not available for {group_by}. Skipping")
                    continue

            else:
                stats = list()
                for file_com in combination:
                    try:
                        prev_stats = self.all_stats_per_format[file_com]
                    except KeyError:
                        raise KeyError(
                            f"Trying to add to combination {file_com}, but does not exists"
                        )
                    else:
                        stats.append(prev_stats)
                stat_df = pd.concat(stats, axis=0, sort=False)

            self.all_stats_per_format[file_base] = stat_df
            self.all_hist_per_format[file_base] = all_hist
            _logger.debug("Done with statistics")

    def write_data(self):
        """write the combined data frame to sqlite lite"""

        count_per_lower_col = Counter([col.lower() for col in self.dataframe.columns])
        for col_lower, multiplicity in count_per_lower_col.items():
            if multiplicity > 1:
                for col in self.dataframe.columns:
                    if col.lower() == col_lower:
                        _logger.info(f"Dropping duplicated column {col}")
                        self.dataframe.drop([col], axis=1, inplace=True)
                        break

        output_file_name = self.cache_file.with_suffix(".sqlite")
        _logger.info(f"Writing dataframe to {output_file_name}")
        with sqlite3.connect(str(output_file_name)) as connection:
            self.dataframe.to_sql(name="dataframe", con=connection, if_exists="replace")

    def read_data(self):
        if not self.cache_file.exists() or self.reset == 0:

            index_name = self.be_id
            _logger.info(f"Reading table data from {self.records_cache_info.file_name}")
            records = read_tables_from_sqlite(
                self.records_cache_info.file_name,
                self.records_cache_info.table_names,
                index_name,
            )

            table_names = ["report", "scoring", "status", "results"]
            index_name = "index"
            _logger.info(
                f"Reading tables {table_names} from {self.internet_nl_filename}"
            )
            tables = read_tables_from_sqlite(
                self.internet_nl_filename, table_names, index_name
            )
            _logger.info(f"Done")
            tables.reset_index(inplace=True)
            tables.rename(columns=dict(index=self.url_key), inplace=True)

            # na are always set to 0
            # some urls did not give any results, drop the empty lines. note that
            # the first column is left out as that is the url name
            tables.dropna(axis=0, how="all", subset=tables.columns[1:], inplace=True)

            # split the tables in numerical and non-numerical part and fill with either NA (for
            # numerical) or with 'nan' for strings.
            original_columns = tables.columns
            tables_num = tables.select_dtypes(include="number")
            number_columns = tables_num.columns
            non_number_columns = [
                col for col in original_columns if col not in number_columns
            ]
            tables_non_num = tables[non_number_columns]

            # fill with NA for numerical values and 'nan' for non-numerical values
            tables_num = tables_num.fillna(pd.NA)
            tables_non_num = tables_non_num.fillna("nan")

            # put back to one data frame again and set order equal to orignal dataframe
            tables = pd.concat([tables_num, tables_non_num], axis=1)
            tables = tables[original_columns]

            self.translations["nans"] = dict(nan=0)

            if self.translations is not None:
                tables = fill_booleans(
                    tables,
                    translations=self.translations.copy(),
                    variables=self.variables,
                )

            rename_all_variables(tables, self.variables)

            for column in tables:
                try:
                    var_props = self.variables.loc[column, :]
                except KeyError:
                    _logger.debug(f"Column {column} not defined in settings. Skipping")
                    continue

                var_type = var_props.get("type")
                var_translate = var_props.get("translateopts")

                if var_translate is not None:
                    # op deze manier kunnen we de vertaling {Nee: 0, Ja: 1} op de column waardes los
                    # laten, zodat we alle Nee met 0 en Ja met 1 vervangen
                    trans = yaml.load(str(var_translate), Loader=yaml.Loader)
                    for nan_string in ("na", "nan", "NaN"):
                        if nan_string in trans.keys():
                            # we have added an 'na' option for the translations. Take care of it
                            is_na = tables[column].isna()
                            if is_na.any():
                                # should not happen anymore because of the dropna above
                                _logger.info(
                                    f"Filling {nan_string} with na in {column}"
                                )
                                tables[column] = tables[column].fillna(nan_string)

                    unique_values = set(tables[column].unique())
                    vals_to_translate = set(trans.keys()).intersection(unique_values)
                    missing_values = unique_values.difference(set(trans.keys()))
                    if vals_to_translate:
                        if missing_values:
                            _logger.warning(
                                f"Column {column} misses the translations for "
                                f"{missing_values}. Please update your settings"
                            )
                        _logger.debug(f"Convert for {column} trans keys {trans}")
                        tables[column] = tables[column].map(trans)
                    else:
                        _logger.debug(f"No Convert for {column} trans keys {trans}")

                if var_type == "dict":
                    tables[column] = tables[column].astype("category")
                elif var_type in ("bool", "percentage", "float"):
                    tables[column] = tables[column].astype("float64")

            # Hier gaan we de url name opschonen. Sla eerst de oorspronkelijke url op
            original_url = "_".join([self.url_key, "original"])
            records = pd.concat(
                [records, records[self.url_key].rename(original_url)], axis=1
            )
            tables = pd.concat(
                [tables, tables[self.url_key].rename(original_url)], axis=1
            )

            _logger.info("Start cleaning urls...")

            if _logger.getEffectiveLevel() > logging.DEBUG:
                show_progress = True
            else:
                show_progress = False
            clean_url_cache = self.cache_directory / Path("clean_url_cache.pkl")
            if clean_url_cache.exists():
                _logger.info(f"Reading clean urls from cache {clean_url_cache}")
                with open(clean_url_cache, "rb") as stream:
                    all_clean_urls, all_suffix = pickle.load(stream)
            else:
                all_clean_urls, all_suffix = get_all_clean_urls(
                    urls=records[self.url_key],
                    show_progress=show_progress,
                    cache_directory=self.tld_extract_cache_directory,
                )
                _logger.info(f"Writing clean urls to cache {clean_url_cache}")
                with open(clean_url_cache, "wb") as stream:
                    pickle.dump([all_clean_urls, all_suffix], stream)
            _logger.info("Done!")
            records[self.url_key] = all_clean_urls
            suffix_df = pd.DataFrame(
                index=records.index, data=all_suffix, columns=[self.suffix_key]
            )
            suffix_df_org = suffix_df.rename(
                columns={self.suffix_key: self.suffix_key + "_org"}
            )
            records = pd.concat([records, suffix_df, suffix_df_org], axis=1)
            records.dropna(subset=[self.url_key], axis=0, inplace=True)
            records.reset_index(inplace=True)

            duplicated = tables[self.url_key].duplicated(keep="first")
            tables = tables[~duplicated]
            tables.dropna(subset=[self.url_key], axis=0, inplace=True)
            tables.dropna(how="all", axis=1, inplace=True)

            # hier voegen we nog afgeleide kolommen to
            tables = add_derived_variables(tables, self.variables)

            # Doe een left join omdat meerdere be's dezelfde url kunnen hebben. Dit is sowieso
            # het geval voor holdings. Dan moeten we de score van holdings ook meerdere keren
            # meenemen
            self.dataframe = pd.merge(
                left=records, right=tables, on=self.url_key, how="left"
            )
            self.dataframe.dropna(
                subset=[self.weight_key], axis="index", how="any", inplace=True
            )
            try:
                has_url = self.dataframe["url"].notnull()
            except KeyError as err:
                _logger.warning(err)
            else:
                self.dataframe = self.dataframe[has_url]
            mask = self.dataframe[self.be_id].duplicated()
            self.dataframe = self.dataframe[~mask]
            self.dataframe.set_index(self.be_id, inplace=True, drop=True)
            self.dataframe = clean_all_suffix(
                dataframe=self.dataframe,
                suffix_key=self.suffix_key,
                variables=self.variables,
            )
            _logger.info(
                f"Writing {self.dataframe.index.size} records to "
                f"cache {self.cache_file.absolute()}"
            )
            with open(str(self.cache_file), "wb") as stream:
                self.dataframe.to_pickle(stream)

        else:
            _logger.debug(f"Reading tables from cache {self.cache_file}")
            with open(str(self.cache_file), "rb") as stream:
                self.dataframe = pd.read_pickle(stream)
            _logger.info(
                f"Read {self.dataframe.index.size} records from "
                f"cache {self.cache_file.absolute()}"
            )

        if self.dump_cache_as_sqlite:
            sqlite_cache = self.cache_file.with_suffix(".sqlite")
            dump_data_frame_as_sqlite(dataframe=self.dataframe, file_name=sqlite_cache)


class DomainPlotter:
    def __init__(
        self,
        scan_data,
        scan_data_key=None,
        default_scan=None,
        plot_info=None,
        show_plots=False,
        barh=False,
        image_directory=None,
        cache_directory=None,
        image_type="pdf",
        max_plots=None,
        tex_prepend_path=None,
        statistics=None,
        variables=None,
        cdf_plot=False,
        bar_plot=False,
        cor_plot=False,
        cumulative=False,
        show_title=False,
        breakdown_labels=None,
        translations: dict = None,
        export_highcharts=False,
        highcharts_directory=None,
        correlations=None,
        tex_horizontal_shift=None,
        bovenschrift=True,
        variables_to_plot=None,
        exclude_variables=None,
        force_plots=False,
        latex_files=False,
        years_to_add_to_plot_legend=None,
        module_info=None,
        english=False,
    ):

        self.english = english
        self.scan_data = scan_data
        self.scan_data_key = scan_data_key
        self.default_scan = default_scan
        self.plot_info = plot_info
        self.show_plots = show_plots
        self.barh = barh
        self.max_plots = max_plots
        self.tex_prepend_path = tex_prepend_path
        self.cache_directory = Path(cache_directory)
        self.statistics = statistics
        self.variables = variables
        self.bar_plot = bar_plot
        self.cdf_plot = cdf_plot
        self.cumulative = cumulative
        self.show_title = show_title
        self.translations = translations
        self.correlations = correlations
        self.export_highcharts = export_highcharts
        self.force_plots = force_plots
        self.years_to_add_to_plot_legend = years_to_add_to_plot_legend
        if highcharts_directory is None:
            self.highcharts_directory = Path(".")
        else:
            self.highcharts_directory = Path(highcharts_directory)
        self.variables_to_plot = variables_to_plot
        self.exclude_variables = exclude_variables

        self.image_type = image_type
        self.image_directory = image_directory
        self.breakdown_labels = breakdown_labels

        self.image_info = ImageFileInfo(
            scan_data_key=scan_data_key, cache_directory=self.cache_directory
        )
        self.image_info.read_cache()

        self.make_plots()

        self.image_info.fix_order(self.variables)
        self.image_info.write_cache()

        if latex_files:
            _logger.debug(f"making latex with bovenschrift={bovenschrift}")
            make_latex_overview(
                image_info=self.image_info,
                variables=self.variables,
                image_directory=self.image_directory,
                cache_directory=self.cache_directory,
                image_files=Path("image_files"),
                tex_prepend_path=self.tex_prepend_path,
                tex_horizontal_shift=tex_horizontal_shift,
                bovenschrift=bovenschrift,
                module_info=module_info,
            )

    #
    def get_plot_cache(self, scan_data_key, plot_key, year_key):
        year_label = f"{year_key}"
        cache_directory = "_".join([self.cache_directory.as_posix(), year_label])
        cache_file = make_plot_cache_file_name(
            cache_directory=Path(cache_directory),
            prefix=scan_data_key,
            file_base=plot_key,
        )
        _logger.debug(f"Reading {cache_file}")
        try:
            with open(cache_file, "rb") as stream:
                stats_df_per_year = pickle.load(stream)
        except FileNotFoundError as err:
            if self.scan_data[scan_data_key][year_key].get("data_file") is None:
                _logger.debug("We are skipping this year as the data is not available.")
            else:
                # we missen de pkl file terwijl we wel een data file hebben. Genereer de foutmelding
                _logger.warning(err)
                _logger.warning("Run script with option '--statistics_to_xls'  first")
            stats_df_per_year = None
        return stats_df_per_year

    def make_plots(self):
        _logger.info("Making the plot")

        legend_translates = dict()

        for plot_key, plot_prop in self.plot_info.items():
            if not plot_prop.get("do_it", True):
                _logger.debug(f"Skipping plot {plot_key}")
                continue

            _logger.debug(f"Plotting plot {plot_key}")
            label = plot_prop.get("label", plot_key)
            figsize = plot_prop.get("figsize")
            highcharts_height = plot_prop.get("highcharts_height")

            stat_prop = self.statistics[plot_key]
            scan_data_key = stat_prop.get("scan_data", self.scan_data_key)

            scan_data_per_year = self.scan_data[scan_data_key]
            last_year = list(scan_data_per_year.keys())[-1]
            scan_data_analyses = scan_data_per_year[last_year]["analyses"]
            variables = scan_data_analyses.variables
            try:
                report_number_empty = variables["report_number"].isna()
            except KeyError:
                report_number_empty = True
                variables["report_number"] = False
            else:
                variables.loc[report_number_empty, "report_number"] = False

            module_info = scan_data_analyses.module_info

            stats_df_per_year = {}
            last_year = None
            df_index_names = None
            for year, scan_info in scan_data_per_year.items():
                df = self.get_plot_cache(
                    scan_data_key=scan_data_key, plot_key=plot_key, year_key=year
                )
                year_label = scan_info.get("label", year)
                if df is not None:
                    stats_df_per_year[year_label] = df
                    last_year = year
                    df_index_names = list(df.index.names)

            if not self.english:
                jaar_level_name = "Jaar"
            else:
                jaar_level_name = "Year"
            index_names = [jaar_level_name] + df_index_names
            new_index_names = df_index_names + [jaar_level_name]
            module_level_name = new_index_names[0]
            question_level_name = new_index_names[1]
            stats_df = pd.concat(stats_df_per_year, names=index_names)
            # zet module vraag optie jaar als volgorde.
            stats_df = stats_df.reorder_levels(new_index_names)

            highcharts_title = plot_prop.get("title")
            export_svg_cdf = False
            export_svg_bar = False
            export_highcharts_cdf = self.export_highcharts
            export_highcharts_bar = self.export_highcharts
            highcharts_directory_cdf = None
            highcharts_directory_bar = None
            cdf_variables = {}
            if self.cdf_plot:
                plot_cdf = plot_prop.get("cdf_plot")
                if isinstance(plot_cdf, dict):
                    cdf_variables = plot_cdf["variables"][scan_data_key]
                    highcharts_title = cdf_variables.get("title")
                export_svg_cdf = False
                if plot_cdf:
                    if cdf_fig_size := plot_cdf.get("figsize"):
                        figsize = cdf_fig_size
            tex_horizontal_shift = None
            if self.bar_plot:
                plot_bar = plot_prop.get("bar_plot")
                highcharts_directory_bar = self.highcharts_directory
                if isinstance(plot_bar, dict):
                    if hc_sub_dir := plot_bar.get("highcharts_output_directory"):
                        highcharts_directory_bar = highcharts_directory_bar / Path(
                            hc_sub_dir
                        )
                    export_svg_bar = plot_bar.get("export_svg", False)
                    export_hc_bar = plot_bar.get("export_highcharts")
                    tex_horizontal_shift = get_windows_or_linux_value(
                        plot_bar.get("tex_horizontal_shift")
                    )
                    plot_bar = plot_bar.get("apply", True)
                    if export_hc_bar is not None:
                        export_highcharts_cdf = export_hc_bar
            else:
                plot_bar = False

            y_max_pdf_plot = plot_prop.get("y_max_pdf_plot", 10)
            y_spacing_pdf_plot = plot_prop.get("y_spacing_pdf_plot", 5)
            y_max_bar_plot = plot_prop.get("y_max_bar_plot")
            legend_position = plot_prop.get("legend_position")
            legend_max_columns = plot_prop.get("legend_max_columns")
            y_spacing_bar_plot = plot_prop.get("y_spacing_bar_plot")
            bar_width = plot_prop.get("bar_width")

            box_margin = plot_prop.get("box_margin")

            sort_values = plot_prop.get("sort_values", False)
            subplot_adjust = plot_prop.get("subplot_adjust")
            reference_lines = plot_prop.get("reference_lines")
            if reference_lines is not None:
                for ref_key, ref_prop in reference_lines.items():
                    stat_prop = self.statistics[ref_key]
                    scan_data_key = stat_prop.get("scan_data", self.default_scan)
                    ref_stat = self.get_plot_cache(
                        scan_data_key=scan_data_key,
                        plot_key=plot_key,
                        year_key=last_year,
                    )
                    reference_lines[ref_key]["data"] = ref_stat

            if plot_prop.get("use_breakdown_keys", False):
                breakdown = self.breakdown_labels[plot_key]
                renames = {v: k for k, v in breakdown.items()}
                stats_df.rename(columns=renames, inplace=True)

            _logger.info(f"Plotting {plot_key}")

            plot_count = 0
            stop_plotting = False
            if stats_df is not None:

                for module_name, module_df in stats_df.groupby(
                    level=module_level_name, sort=False
                ):
                    do_this_module = True
                    for mod_key, mod_prop in module_info.items():
                        if mod_prop.get("label") == module_name and not mod_prop.get(
                            "include", True
                        ):
                            do_this_module = False
                    if not do_this_module:
                        continue

                    _logger.info(f"Module {module_name}")
                    if stop_plotting:
                        break
                    for question_name, question_df in module_df.groupby(
                        level=question_level_name, sort=False
                    ):
                        _logger.debug(f"Question {question_name}")

                        # voorlaatste kolom bevat de variabele namen
                        variable_name_key = question_df.index.names[-2]
                        plot_variable = question_df.index.get_level_values(
                            variable_name_key
                        ).values[0]
                        original_name = re.sub(r"_\d\.0$", "", plot_variable)
                        question_type = variables.loc[original_name, "type"]
                        unit = variables.loc[original_name, "unit"]
                        keep_options = variables.loc[original_name, "keep_options"]
                        section = variables.loc[original_name, "section"]
                        question_df_clean = question_df.droplevel(variable_name_key)

                        # variables_to_plot wordt als een list van een list in een tuple meegegeven
                        # dus ([[variable1], [variables2]). Haal eerst level 0 eruit om te tuple
                        # te verwijderen. Als variable_to_plot niet gegeven is dan is deze waarde
                        # None, en slaan we het over. Als hij wel gegeven is dan zetten we de list
                        # van lists om in een platte list
                        if self.variables_to_plot is not None:
                            var_to_plot_clean = [
                                vv[0] for vv in self.variables_to_plot if vv is not None
                            ]
                            if original_name not in var_to_plot_clean:
                                _logger.debug(
                                    f"{original_name} not in variables to plot {self.variables_to_plot}. "
                                    f"Skipping..."
                                )
                                continue

                        if self.exclude_variables is not None:
                            exclude_vars_clean = [
                                vv[0] for vv in self.exclude_variables if vv is not None
                            ]
                            if original_name in exclude_vars_clean:
                                _logger.debug(
                                    f"{original_name} in exclude variables {self.exclude_variables}. "
                                    f"Skipping..."
                                )
                                continue

                        plot_info = PlotInfo(
                            variables_df=variables,
                            var_name=original_name,
                            breakdown_name=plot_key,
                        )
                        export_highcharts = export_highcharts_bar
                        if cdf_prop := cdf_variables.get(original_name):
                            highcharts_directory_cdf = self.highcharts_directory
                            if highcharts_info_per_year := cdf_prop.get(
                                "highcharts_info_per_year"
                            ):
                                for (
                                    hc_year_key,
                                    hc_year_prop,
                                ) in highcharts_info_per_year.items():
                                    hc_dir = highcharts_directory_cdf / Path(
                                        hc_year_prop["highcharts_directory"]
                                    )
                                    hc_lab = hc_year_prop.get("highcharts_label")
                                    highcharts_info_per_year[hc_year_key] = dict(
                                        highcharts_directory=hc_dir,
                                        highcharts_label=hc_lab,
                                    )
                            export_svg_cdf = cdf_prop.get("export_svg", False)
                            export_hc_cdf = cdf_prop.get("export_highcharts")
                            plot_cdf = cdf_prop.get("apply", True)
                            if export_hc_cdf is not None:
                                export_highcharts_cdf = export_hc_cdf
                        else:
                            plot_cdf = False
                            highcharts_info_per_year = None
                        if plot_info.directory is not None:
                            # we overschrijven hier de subdir die onder de statistiek opgegeven is
                            highcharts_directory = (
                                self.highcharts_directory / plot_info.directory
                            )
                        else:
                            if plot_bar:
                                highcharts_directory = highcharts_directory_bar
                            else:
                                highcharts_directory = highcharts_directory_cdf
                        if plot_info.label is not None:
                            title = plot_info.label
                        else:
                            title = highcharts_title

                        if title is not None:
                            title = re.sub("\s{2,}", " ", title)

                        if plot_info.y_max is not None:
                            y_max = plot_info.y_max
                        else:
                            y_max = y_max_bar_plot

                        if plot_info.legend_position is not None:
                            legend_pos = plot_info.legend_position
                        else:
                            legend_pos = legend_position

                        if plot_info.y_spacing is not None:
                            y_spacing = plot_info.y_spacing
                        else:
                            y_spacing = y_spacing_bar_plot

                        if plot_info.bar_width is not None:
                            bar_width = plot_info.bar_width
                        else:
                            bar_width = bar_width

                        if keep_options:
                            # als keep options gegeven is dan houden we alle opties
                            valide_opties = variables.loc[
                                original_name, "options"
                            ].values()
                            mask = get_option_mask(
                                question_df=question_df_clean,
                                variables=variables,
                                question_type=question_type,
                                valid_options=valide_opties,
                            )
                            plot_df = module_df.loc[
                                (module_name, question_name, mask)
                            ].copy()

                        else:
                            # neem de default die we als true bestempelen
                            valide_opties = None
                            mask = get_option_mask(
                                question_df=question_df_clean,
                                variables=variables,
                                question_type=question_type,
                                valid_options=valide_opties,
                            )

                            plot_df = question_df_clean.loc[
                                (module_name, question_name, mask)
                            ].copy()

                        # dit is niet meer nodig omdat de kleuren toch gelijk blijven
                        # plot_df = add_missing_years(plot_df,
                        #                            years_to_plot=self.years_to_add_to_plot_legend,
                        #                            jaar_level_name=jaar_level_name,
                        #                            column=original_name)

                        if variables.loc[original_name, "report_number"]:
                            normalize_data = True
                        else:
                            normalize_data = False

                        if self.translations is not None:
                            plot_df.rename(columns=self.translations, inplace=True)

                        xoff = 0
                        yoff = 0

                        if reference_lines is not None:
                            for ref_key, ref_prop in reference_lines.items():
                                ref_stat_df = reference_lines[ref_key]["data"]
                                ref_quest_df = None
                                for ref_quest_name, ref_quest_df in ref_stat_df.groupby(
                                    level=question_level_name
                                ):
                                    if ref_quest_name == question_name:
                                        break
                                if ref_quest_df is not None:
                                    mask2 = get_option_mask(
                                        question_df=ref_quest_df,
                                        variables=variables,
                                        question_type=question_type,
                                    )
                                    try:
                                        ref_df = ref_quest_df.loc[
                                            (module_name, question_name, mask2)
                                        ].copy()
                                    except KeyError as err:
                                        _logger.warning(err)
                                    else:
                                        reference_lines[ref_key]["plot_df"] = ref_df

                        _logger.info(f"Plot nr {plot_count}")
                        if plot_bar:
                            if keep_options:
                                for (
                                    year_key,
                                    local_scan_info,
                                ) in scan_data_per_year.items():
                                    year = local_scan_info.get("label", year_key)
                                    this_year_df = plot_df.loc[
                                        slice(None),
                                        slice(None),
                                        slice(None),
                                        slice(None),
                                        year,
                                    ]
                                    image_file = make_bar_plot_stacked(
                                        year=year_key,
                                        plot_df=this_year_df,
                                        plot_key=plot_key,
                                        plot_variable=plot_variable,
                                        scan_data_key=scan_data_key,
                                        module_name=module_name,
                                        question_name=question_name,
                                        image_directory=self.image_directory,
                                        show_plots=self.show_plots,
                                        figsize=figsize,
                                        image_type=self.image_type,
                                        reference_lines=reference_lines,
                                        xoff=xoff,
                                        yoff=yoff,
                                        show_title=self.show_title,
                                        barh=self.barh,
                                        subplot_adjust=subplot_adjust,
                                        box_margin=box_margin,
                                        sort_values=sort_values,
                                        y_max_bar_plot=y_max,
                                        y_spacing_bar_plot=y_spacing,
                                        translations=self.variables.loc[
                                            original_name, "options"
                                        ],
                                        export_highcharts=export_highcharts,
                                        export_svg=export_svg_bar,
                                        highcharts_directory=highcharts_directory,
                                        title=title,
                                        legend_position=legend_pos,
                                        normalize_data=normalize_data,
                                        force_plot=self.force_plots,
                                        enable_highcharts_legend=plot_info.enable_highcharts_legend,
                                        unit=unit,
                                        english=self.english,
                                    )
                            else:
                                image_file = make_bar_plot(
                                    plot_df=plot_df,
                                    plot_key=plot_key,
                                    plot_variable=plot_variable,
                                    scan_data_key=scan_data_key,
                                    module_name=module_name,
                                    question_name=question_name,
                                    image_directory=self.image_directory,
                                    show_plots=self.show_plots,
                                    figsize=figsize,
                                    highcharts_height=highcharts_height,
                                    image_type=self.image_type,
                                    reference_lines=reference_lines,
                                    xoff=xoff,
                                    yoff=yoff,
                                    show_title=self.show_title,
                                    barh=self.barh,
                                    subplot_adjust=subplot_adjust,
                                    box_margin=box_margin,
                                    sort_values=sort_values,
                                    y_max_bar_plot=y_max,
                                    y_spacing_bar_plot=y_spacing,
                                    translations=self.translations,
                                    export_highcharts=export_highcharts,
                                    export_svg=export_svg_bar,
                                    highcharts_directory=highcharts_directory,
                                    title=title,
                                    legend_position=legend_pos,
                                    legend_max_columns=legend_max_columns,
                                    normalize_data=normalize_data,
                                    force_plot=self.force_plots,
                                    enable_highcharts_legend=plot_info.enable_highcharts_legend,
                                    unit=unit,
                                    english=self.english,
                                    bar_width=bar_width,
                                )

                                _logger.debug(
                                    f"Store [{original_name}][{label}] : {image_file}"
                                )
                                self.image_info.add_entry(
                                    plot_key=plot_key,
                                    plot_info=self.plot_info,
                                    image_key=original_name,
                                    section=section,
                                    file_name=image_file,
                                    sub_image_label=label,
                                    tex_right_shift=tex_horizontal_shift,
                                )

                        if plot_cdf:
                            for year in scan_data_per_year.keys():
                                try:
                                    scan_data_analyses_year = scan_data_per_year[year][
                                        "analyses"
                                    ]
                                except KeyError:
                                    _logger.info(
                                        f"Year {year} does not have data. Skipping"
                                    )
                                    continue
                                hist_info = scan_data_analyses_year.all_hist_per_format[
                                    plot_key
                                ][original_name]
                                highcharts_info = highcharts_info_per_year[year]

                                if hist_info is not None and isinstance(
                                    hist_info, dict
                                ):
                                    for grp_key, hist in hist_info.items():
                                        if hist is None:
                                            _logger.warning(
                                                f"Hist {grp_key} does not have a histogram. Skipping"
                                            )
                                            continue
                                        im_file_2 = make_cdf_plot(
                                            hist=hist,
                                            plot_key=plot_key,
                                            scan_data_key=scan_data_key,
                                            grp_key=grp_key,
                                            module_name=module_name,
                                            question_name=question_name,
                                            image_file_base=original_name,
                                            image_directory=self.image_directory,
                                            show_plots=self.show_plots,
                                            figsize=figsize,
                                            image_type=self.image_type,
                                            reference_lines=reference_lines,
                                            cummulative=self.cumulative,
                                            xoff=xoff,
                                            yoff=yoff,
                                            y_max=y_max_pdf_plot,
                                            y_spacing=y_spacing_pdf_plot,
                                            translations=self.translations,
                                            export_highcharts=export_highcharts_cdf,
                                            export_svg=export_svg_cdf,
                                            highcharts_info=highcharts_info,
                                            title=title,
                                            year=year,
                                        )
                            if self.show_plots:
                                plt.show()

                        plot_count += 1
                        if self.max_plots is not None and plot_count == self.max_plots:
                            _logger.info(
                                f"Maximum number of plot ({self.max_plots}) reached"
                            )
                            stop_plotting = True
                            break


class PlotInfo:
    def __init__(self, variables_df, var_name, breakdown_name):
        self.variables_df = variables_df
        self.var_name = var_name
        self.breakdown_name = breakdown_name

        self.label = None
        self.directory = None
        self.y_max = None
        self.y_spacing = None
        self.legend_position = None
        self.enable_highcharts_legend = True
        self.bar_width = None

        self.get_plot_info()

    def get_plot_info(self):
        """In de variables dataframe kunnen we ook uitdrukkelijk de highcharts directory en highcharts
        label opgeven per variabele. Zoek dat hier op"""
        label = None
        directory = None
        try:
            var_prop = self.variables_df.loc[self.var_name]
        except KeyError:
            _logger.debug(
                f"could not find variable {self.var_name} in variables dataframe"
            )
        else:
            info_per_breakdown = var_prop["info_per_breakdown"]
            if info_per_breakdown is not None:
                try:
                    info = info_per_breakdown[self.breakdown_name]
                except KeyError:
                    _logger.debug(
                        f"variable {self.var_name} does not have a breakdown defined"
                    )
                else:
                    self.directory = info.get("highcharts_directory")
                    if self.directory is not None:
                        self.directory = Path(self.directory)
                    self.label = info.get("highcharts_label")
                    self.y_max = info.get("y_max")
                    self.y_spacing = info.get("y_spacing")
                    self.bar_width = info.get("bar_width")
                    self.legend_position = get_windows_or_linux_value(
                        info.get("legend_position")
                    )
                    self.enable_highcharts_legend = info.get(
                        "enable_highcharts_legend", True
                    )


def add_missing_years(plot_df, years_to_plot=None, jaar_level_name="Jaar", column=None):
    """
    Voeg missende jaren toe

    Args:
        plot_df: pd.DataFrame
            DataFrame om te plotetn
        years_to_plot: list
            De jaren die we willen plotten
        jaar_level_name: str
            De naam van de level= van de jaren
        column: str
            Naam van de column voor de foutmelding

    Returns:
        pd.DataFrame
    """

    years_in_plot = plot_df.index.get_level_values(jaar_level_name)
    missing_years = set(years_to_plot).difference(years_in_plot)
    if missing_years:
        index_names = plot_df.index.names
        df = plot_df.reset_index().set_index(jaar_level_name)
        try:
            df = df.reindex(years_to_plot)
        except ValueError as err:
            _logger.warning(f"{err}. Check  {column}")
        for column_name in index_names:
            if column_name == jaar_level_name:
                continue
            df[column_name] = df[column_name].pad()
        plot_df = df.reset_index().set_index(index_names, drop=True)

    return plot_df


def calculate_histogram_per_breakdown(
    data: DataFrame, var_key: str, df_weights: Series, n_bins: int = 100
) -> dict:
    """
    Bereken per breakdown van de data het histogram die hoort bij var_key

    Parameters
    ----------
    data: DataFrame
        De data met breakdown op de index
    var_key: str
        De naam van de kolom waarvoor we de histogram gaan berekenen
    df_weights: Series
        De weegfactoren die we voor de histogram gebruiken
    n_bins: int
        Aan binnen in het histogram

    Returns
    -------
    dict:
        De histogrammen per breakdown


    """

    histogram_per_breakdown = dict()

    for grp_key, df in data.groupby(level=0):
        # initieer histogram voor deze breakdown met None
        histogram_per_breakdown[grp_key] = None

        try:
            ww = df_weights.loc[grp_key, "ratio_units"].to_numpy()
        except KeyError:
            _logger.debug("Could not get weight factors. Skip for now")
            continue
        try:
            dd = df.loc[grp_key, var_key].to_numpy()
        except KeyError:
            _logger.debug(f"Could not get data belonging to {var_key}. Skip for now")
            continue

        try:
            histogram = np.histogram(
                dd,
                weights=ww,
                density=False,
                bins=n_bins,
                range=(0, 100),
            )
        except ValueError as err:
            _logger.debug("Fails for dicts. Skip for now")
        else:
            _logger.debug(f"Success with {var_key}")
            histogram_per_breakdown[grp_key] = histogram

    return histogram_per_breakdown



import logging
from pathlib import Path

from pylatex import Document, Figure, NoEscape, Command, SubFigure
from pylatex.base_classes import Environment, CommandBase, Arguments
from pylatex.package import Package

_logger = logging.getLogger(__name__)


class ExampleEnvironment(Environment):
    """
    A class representing a custom LaTeX environment.

    This class represents a custom LaTeX environment named
    ``exampleEnvironment``.
    """

    _latex_name = "exampleEnvironment"
    packages = [Package("mdframed")]


class SubFloat(CommandBase):
    """
    A class representing a custom LaTeX command.

    This class represents a custom LaTeX command named
    ``exampleCommand``.
    """

    _latex_name = "subfloat"


def make_latex_overview(
    image_info=None,
    variables=None,
    image_directory=None,
    cache_directory=None,
    image_files=None,
    tex_horizontal_shift="-2cm",
    tex_prepend_path=None,
    bovenschrift=False,
    module_info=None,
):
    """
    Maak latex output file met alle plaatjes

    Args:
        module_info: class
            Informatie van de modules
        cache_directory:  obj:Path
        image_info: object: ImageInfo
        variables: dict met variabele eigenschappen
        image_directory: str
        image_files: obj:Path
        tex_prepend_path: str
        tex_horizontal_shift: verschuiving naar links
        bovenschrift: boolean
            Voeg caption bovenaan figuren
    """
    if tex_prepend_path is None:
        full_image_directory = image_directory
    else:
        full_image_directory = tex_prepend_path / image_directory

    doc_per_module = dict()

    for original_name, image_prop in image_info.data.items():
        _logger.debug(f"Adding {original_name}")
        caption = variables.loc[original_name, "label"]
        module = variables.loc[original_name, "module"]
        section_key = variables.loc[original_name, "section"]
        section_title = None
        if section_key:
            try:
                sections_for_module = module_info[module]["sections"]
            except KeyError as err:
                _logger.warning(
                    f"{err}\n"
                    f"You have added a section {section_key} to module {module},"
                    f"but no sections can be found under 'module_info'. Please"
                    f"specify"
                )
            else:
                try:
                    section_prop = sections_for_module[section_key]
                except KeyError as err2:
                    _logger.warning(
                        f"{err2}\nNo section key {section_key} was found. skipping"
                    )
                else:
                    section_title = section_prop["title"]
        try:
            doc = doc_per_module[module]
        except KeyError:
            doc = Document(default_filepath=full_image_directory)
            doc_per_module[module] = doc
        if section_title is not None:
            doc.append(Command("section", section_title))
            doc.append(Command("label", NoEscape("sec:" + section_key)))
        with doc.create(Figure(position="htb")) as plots:
            add_new_line = True
            if bovenschrift:
                # hiermee worden caption boven toegevoegd, maar ik wil hier vanaf gaan wijken.
                plots.add_caption(caption)
                ref = "_".join([image_info.scan_data_key, original_name])
                ref_label = Command("label", NoEscape("fig:" + ref))
                plots.append(ref_label)
            for sub_image_key, sub_image_prop in image_prop.items():
                sub_image_label = sub_image_prop.get("sub_image_label", "Empty")
                image_name = sub_image_prop["file_name"]
                horizontal_shift = sub_image_prop.get(
                    "tex_right_shift", tex_horizontal_shift
                )
                with doc.create(SubFigure(width=NoEscape(r"\linewidth"))) as sub_plot:
                    if tex_prepend_path is None:
                        full_image_name = Path(image_name)
                    else:
                        if image_name is not None:
                            full_image_name = tex_prepend_path / image_name
                        else:
                            full_image_name = Path("bla")
                    _logger.debug(f"Adding {full_image_name}")
                    ref = "_".join(
                        [
                            image_info.scan_data_key,
                            original_name,
                            sub_image_label.lower().replace(" ", "_"),
                        ]
                    )
                    ref_sublabel = Command("label", NoEscape("fig:" + ref))
                    lab = Command("footnotesize", Arguments(sub_image_label))
                    include_graphics = Command(
                        "includegraphics",
                        NoEscape(full_image_name.with_suffix("").as_posix()),
                    )
                    if horizontal_shift is not None:
                        include_graphics = Command(
                            "hspace",
                            Arguments(NoEscape(horizontal_shift), include_graphics),
                        )
                    # sub_plot = SubFloat(
                    #    options=[lab],
                    #    arguments=Arguments(hspace))
                    if not bovenschrift:
                        sub_plot.append(include_graphics)
                    sub_plot.add_caption(lab)
                    sub_plot.append(ref_sublabel)
                    if bovenschrift:
                        sub_plot.append(include_graphics)
                if add_new_line:
                    plots.append(Command("newline"))
                    add_new_line = False
            if not bovenschrift:
                # voeg een onderschrift toe
                plots.add_caption(caption)
                ref_label = Command("label", NoEscape("fig:" + original_name))
                plots.append(ref_label)

    for module, doc in doc_per_module.items():
        file_name = cache_directory / Path(
            "_".join(
                [
                    image_info.scan_data_key,
                    image_files.with_suffix("").as_posix(),
                    module,
                ]
            )
        )
        _logger.info(f"Writing tex file list to {file_name}.tex")
        doc.generate_tex(filepath=file_name.as_posix())
        file_name = file_name.with_suffix(".tex")
        new_lines = list()
        start = False
        with open(file_name.as_posix(), "r") as stream:
            for line in stream.readlines():
                if "figure" in line or "section" in line:
                    start = True
                if "end{document}" in line:
                    start = False
                if start:
                    new_lines.append(line)
        with open(file_name.as_posix(), "w") as stream:
            stream.writelines(new_lines)



import logging
import sqlite3
import sys
from pathlib import Path

import pandas as pd
import yaml
from tqdm import tqdm

from internetnl_scan.utils import get_clean_url

from weighted_sample_statistics import reorganise_stat_df

_logger = logging.getLogger(__name__)
tld_logger = logging.getLogger("tldextract")
tld_logger.setLevel(logging.WARNING)


# noinspection SqlDialectInspection
def read_tables_from_sqlite(filename: Path, table_names, index_name) -> pd.DataFrame:
    if isinstance(table_names, str):
        table_names = [table_names]

    if not filename.exists():
        _logger.warning(
            "Records file not found. Make sure you set the environment variable "
            "RECORDS_CACHE_DIR_<yearkey> for the file and RECORDS_TABLE_RECS_<yearkey> and "
            "RECORDS_TABLE_INFO<yearkey> for the tables or pass it via the command line argument "
            "--records_cache_dir"
        )
        raise FileNotFoundError(f"Records file not found {filename.absolute()}")

    _logger.info(f"Reading from {filename}")
    connection = sqlite3.connect(filename.as_posix())
    tables = list()
    for table_name in table_names:
        _logger.debug(f"Reading table {table_name}")
        df = pd.read_sql(
            f"select * from {table_name}", con=connection, index_col=index_name
        )
        tables.append(df)
    _logger.debug(f"Done")
    if len(tables) > 1:
        tables_df = pd.concat(tables, axis=1)
    else:
        tables_df = tables[0]

    _logger.debug(f"Closing database")

    connection.close()

    _logger.debug(f"Done reading")
    return tables_df


def add_derived_variables(tables, variables):
    """
    Add the variables we defined in the settings files which do not exist yet, but are defined with an eval statement

    Args:
        tables: pd.DataFrame
            original table of variables
        variables: pd.DataFrame
            properties of variables

    Returns:
        pd.DataFame

    """
    undefined_variables = variables.index.difference(tables.columns)

    for variable_name in undefined_variables:
        _logger.debug(f"deriving properties for {variable_name}")

        var_props = variables.loc[variable_name, :]
        eval_statement = var_props.get("eval")
        if eval_statement is None:
            _logger.debug(
                f"Column {variable_name} does not exist but settings do not provide an eval statement.\n"
                f"Is ok, not all years have all properties defined. Only if an eval statement"
                f"is defined, we are going to add a new columns"
            )
            continue

        _logger.info(f"creating new column {variable_name} as {eval_statement}")
        tables[variable_name] = tables.eval(eval_statement)

    return tables


def fill_booleans(tables, translations, variables):
    for col in tables.columns:
        convert_to_bool = True
        try:
            var_type = variables.loc[col, "type"]
        except KeyError:
            pass
        else:
            if var_type == "dict":
                convert_to_bool = False

        if convert_to_bool:
            unique_values = tables[col].unique()
            if len(unique_values) <= 3:
                for trans_key, trans_prop in translations.items():
                    bool_keys = set(trans_prop.keys())
                    intersection = bool_keys.intersection(unique_values)
                    if intersection:
                        nan_val = set(unique_values).difference(bool_keys)
                        # if nan_val:
                        #    trans_prop[list(nan_val)[0]] = np.nan
                        for key, val in trans_prop.items():
                            mask = tables[col] == key
                            if any(mask):
                                tables.loc[mask, col] = float(val)
    return tables


def prepare_stat_data_for_write(
    all_stats,
    file_base,
    variables,
    module_key,
    variable_key,
    breakdown_labels=None,
    n_digits=3,
    connection=None,
):
    data = pd.DataFrame.from_dict(all_stats)
    if connection is not None:
        data.to_sql(name=file_base, con=connection, if_exists="replace")

    stat_df = reorganise_stat_df(
        records_stats=data,
        variables=variables,
        use_original_names=True,
        module_key=module_key,
        variable_key=variable_key,
        n_digits=n_digits,
        sort_index=False,
    )
    index_names = list(stat_df.index.names)
    new_index_names = index_names + [stat_df.columns[0]]
    stat_df = stat_df.reset_index().set_index(new_index_names, drop=True)
    if breakdown_labels is not None:
        try:
            labels = breakdown_labels[file_base]
        except KeyError:
            _logger.info(f"No breakdown labels for {file_base}")
        else:
            stat_df.rename(columns=labels, inplace=True)

    return stat_df


def get_option_mask(question_df, variables, question_type, valid_options=None):
    """get the mask to filter the positive options from a question"""
    mask_total = None
    if valid_options is None:
        options = ("Passed", "Yes", "Good")
    else:
        options = valid_options
    if question_type == "dict":
        for optie in options:
            mask = question_df.index.get_level_values(2) == optie
            if mask_total is None:
                mask_total = mask
            else:
                mask_total = mask | mask_total
    else:
        mask_total = pd.Series(True, index=question_df.index)

    return mask_total


def impose_variable_defaults(
    variables, module_info: dict = None, module_key: str = None
):
    """
    Impose default values to  the variables data frame

    Parameters
    ----------
    variables: pd.DataFrame
        Dataframe with the initial variables
    module_info: pd.DataFrame
        Dataframe with information per module
    module_key: str
        Key of the module in the dataframe

    Returns
    -------
    pd.DataFrame
        Filled dataframe

    """
    variables["type"] = "bool"
    variables["section"] = ""
    variables["fixed"] = False
    variables["original_name"] = variables.index.values
    variables["label"] = ""
    variables["question"] = ""
    variables["module_label"] = ""
    variables["module_include"] = True
    # if the check  flag is true , it indicates this is a check question which can be discarded
    # in the final output
    variables["check"] = False
    variables["optional"] = False
    variables["no_impute"] = False
    variables["info_per_breakdown"] = None

    variables["gewicht"] = "units"
    variables["keep_options"] = False
    variables["eval"] = None
    variables["unit"] = None

    # als toevallig de eerste key: value in de options een dict is dan kan je geen from_dict
    # gebruiken. Daarom voegen we nu een dummy string to, die halen we dadelijk weer weg
    dummy = "dummy"
    options = {dummy: dummy}
    filter_dummy = {dummy: dummy}
    translate = {dummy: dummy}
    for var_key, var_row in variables.iterrows():
        var_prop = var_row["properties"]
        # loop over the given column names as try to read the value from 'properties' field
        # in the variables dataframe. This properties field is the dict we have read from
        # the settings file which may contain the same key  with a value. If this value was
        # defined for the current variable, copy it to the associate column in the data frame
        # such that we can access it more easily
        for name in (
            "type",
            "fixed",
            "original_name",
            "question",
            "label",
            "check",
            "optional",
            "gewicht",
            "no_impute",
            "info_per_breakdown",
            "report_number",
            "section",
            "keep_options",
            "eval",
            "unit",
        ):
            try:
                variables.loc[var_key, name] = var_prop[name]
            except ValueError:
                # de info_per_breakdown is een dictionary die we met 'at' moeten imposen
                variables.at[var_key, name] = var_prop.get(name)
            except KeyError:
                pass
        # separately get the options field as that contains a dict and therefore can not be
        # imposed as a single value to a row. Instead, collect them and append as a single col
        try:
            var_options = var_prop["options"]
        except KeyError:
            options[var_key] = None
        else:
            options[var_key] = var_options

        try:
            var_filter = var_prop["filter"]
        except KeyError:
            filter_dummy[var_key] = None
        else:
            filter_dummy[var_key] = var_filter

        try:
            var_translate = var_prop["translate"]
        except KeyError:
            translate[var_key] = None
        else:
            translate[var_key] = var_translate

        # add the module label  to this dataframe as well
        if module_info is not None:
            module_name = var_row[module_key]
            try:
                module_label = module_info[module_name]["label"]
            except KeyError:
                _logger.warning("failed to get the label from {}".format(module_name))
            else:
                variables.loc[var_key, "module_label"] = module_label
            try:
                module_include = module_info[module_name]["include"]
            except KeyError:
                _logger.warning(
                    "failed to get the include flag from {}".format(module_name)
                )
            else:
                variables.loc[var_key, "module_include"] = module_include

    # create data frame of one columns from the option dictionaries.
    opt_df = pd.DataFrame.from_dict(options, orient="index").rename(
        columns={0: "options"}
    )
    opt_df = opt_df[opt_df.index != dummy]

    filter_df = pd.DataFrame.from_dict(filter_dummy, orient="index").rename(
        columns={0: "filter"}
    )
    filter_df = filter_df[filter_df.index != dummy]

    trans_df = pd.DataFrame.from_dict(translate, orient="index").rename(
        columns={0: "translateopts"}
    )
    trans_df = trans_df[trans_df.index != dummy]

    # drop the original column with properties
    variables.drop(["properties"], inplace=True, axis=1)

    # merge the options column with the rest of the columns
    variables = pd.concat([variables, opt_df], axis=1)
    variables = pd.concat([variables, filter_df], axis=1)
    variables = pd.concat([variables, trans_df], axis=1)

    # check if we have a dict data type that does not has a option field set. In that case
    # raise a warning: all the dict type need the options defined
    is_dict = variables["type"] == "dict"
    if (is_dict & variables["options"].isnull()).sum() > 0:
        raise ValueError("Found a dict with no options defined")

    _logger.debug("Done")
    return variables


def add_missing_groups(all_stats, group_by, group_by_original, missing_groups):
    new_stats = {}
    for indicator, data_df in all_stats.items():
        for gb_new, gb_org in zip(group_by, group_by_original):
            data_df.index = data_df.index.rename(gb_org)
        if missing_groups is not None:
            df_extra = pd.DataFrame(index=missing_groups, columns=data_df.columns)
            df_extra = df_extra.astype(data_df.dtypes)
            data_df = pd.concat([df_extra, data_df])
            new_stats[indicator] = data_df
    return new_stats


def clean_all_suffix(dataframe, suffix_key, variables):
    """
    Hier gaan we de suffixen selecteren die we gedefinieerd hebben.

    Args:
        dataframe: dataframe met tabellen, waaronder een kolom met website extensies
        suffix_key: de naam van de kolom met website extensies
        variables: dataframe met variable informatie. Moet minimaal een variabele
        gelijk aan de suffix_key hebben waarin de categorieën gedefinieerd zijn
    Returns:
        dataframe

    """

    if suffix_key in variables.index:
        translateopts = variables.loc[suffix_key, "translateopts"]
        categories = dataframe[suffix_key].astype("category")
        # we nemen aan dat de laatste category in de definitie 'rest' is
        categorie_names = list(translateopts.keys())[:-1]
        rest_category = list(translateopts.keys())[-1]
        # we gaan hier categorieen toevoegen op basis van het lijst dat we
        # in de settings file gegeven hebben (.nl, .com, .eu). De laatste beschouwen
        # we als rest categorie waartoe we alle dat niet bij de hoofdcategorieen
        # hoort gaan indelen.
        categories = categories.cat.set_categories(categorie_names)
        # alle velden die nu nog geen categorie hebben worden nu aangeduid
        # met de rest categorie
        categories = categories.cat.add_categories(rest_category)

        # de rest categorieen worden nu op 'na' gezet
        categories.fillna(rest_category, inplace=True)

        # dit is nodig om van de strings 'com', 'nl' etc getallen 1, 2, 3 etc te maken.
        categories = categories.astype(str)
        # vertaal de strings per categorie 'com', 'nl' etc naar de digits
        trans = yaml.load(str(translateopts), Loader=yaml.Loader)
        categories = categories.map(trans)

        # kopieer terug naar data frame als category type
        dataframe[suffix_key] = categories.astype("category")

    else:
        _logger.info("Could not find suffix info to translate")

    return dataframe


def get_all_clean_urls(urls, show_progress=False, cache_directory=None):
    if show_progress:
        progress_bar = tqdm(
            total=urls.size,
            file=sys.stdout,
            position=0,
            ncols=100,
            leave=True,
            colour="GREEN",
        )
    else:
        progress_bar = None

    all_clean_urls = list()
    all_suffix = list()

    for url in urls:
        clean_url, suffix = get_clean_url(url, cache_dir=cache_directory)
        _logger.debug(f"Converted {url} to {clean_url}")
        all_clean_urls.append(clean_url)
        all_suffix.append(suffix)
        if progress_bar:
            if clean_url is not None:
                progress_bar.set_description("{:5s} - {:30s}".format("URL", clean_url))
            else:
                progress_bar.set_description("{:5s} - {:30s}".format("URL", "None"))
            progress_bar.update()
    return all_clean_urls, all_suffix


def get_windows_or_linux_value(value):
    """Pas de waarde aan als deze in een dict gegeven is met een windows en linux veld"""
    if isinstance(value, dict):
        if "win" in sys.platform:
            new_value = value["windows"]
        else:
            new_value = value["linux"]
    else:
        new_value = value

    return new_value


def dump_data_frame_as_sqlite(dataframe, file_name):
    """Dump data als sqlite, maar zorg dat je duplicates eruit haalt"""

    # maak lower van de columnnamen want sqlite is case insensitive
    is_duplicated_column = dataframe.columns.str.lower().duplicated()
    duplicated_columns = dataframe.columns[is_duplicated_column]
    _logger.debug(f"Dropping duplicated columns {duplicated_columns}")
    clean_df = dataframe.drop(columns=duplicated_columns)

    _logger.info(f"Writing cache as sqlite {file_name}")
    with sqlite3.connect(file_name) as connection:
        clean_df.to_sql(name="table", con=connection, if_exists="replace")
